# Copyright 2026 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
from textwrap import dedent
import pandas as pd
from pylegend._typing import (
    TYPE_CHECKING,
    PyLegendDict,
    PyLegendOptional,
    PyLegendSequence,
    PyLegendTypeVar,
    PyLegendUnion
)
from pylegend.core.database.sql_to_string import SqlToStringConfig, SqlToStringFormat
from pylegend.core.language.pandas_api.pandas_api_aggregate_specification import PyLegendAggInput
from pylegend.core.language.pandas_api.pandas_api_series import (
    SupportsToPureExpression,
    SupportsToSqlExpression
)
from pylegend.core.language.pandas_api.pandas_api_tds_row import PandasApiTdsRow
from pylegend.core.language.shared.column_expressions import PyLegendColumnExpression
from pylegend.core.language.shared.expression import (
    PyLegendExpressionBooleanReturn,
    PyLegendExpressionDateReturn,
    PyLegendExpressionDateTimeReturn,
    PyLegendExpressionFloatReturn,
    PyLegendExpressionDecimalReturn,
    PyLegendExpressionIntegerReturn,
    PyLegendExpressionNumberReturn,
    PyLegendExpressionStrictDateReturn,
    PyLegendExpressionStringReturn,
    PyLegendExpression,
)
from pylegend.core.language.shared.helpers import generate_pure_lambda, escape_column_name
from pylegend.core.language.shared.primitives.boolean import PyLegendBoolean
from pylegend.core.language.shared.primitives.date import PyLegendDate
from pylegend.core.language.shared.primitives.datetime import PyLegendDateTime
from pylegend.core.language.shared.primitives.float import PyLegendFloat
from pylegend.core.language.shared.primitives.decimal import PyLegendDecimal
from pylegend.core.language.shared.primitives.integer import PyLegendInteger
from pylegend.core.language.shared.primitives.number import PyLegendNumber
from pylegend.core.language.shared.primitives.primitive import (
    PyLegendPrimitive,
    PyLegendPrimitiveOrPythonPrimitive
)
from pylegend.core.language.shared.primitives.strictdate import PyLegendStrictDate
from pylegend.core.language.shared.primitives.string import PyLegendString
from pylegend.core.sql.metamodel import Expression, QuerySpecification, SingleColumn, QualifiedNameReference, QualifiedName
from pylegend.core.tds.abstract.frames.base_tds_frame import BaseTdsFrame
from pylegend.core.tds.pandas_api.frames.functions.rank_function import RankFunction
from pylegend.core.tds.pandas_api.frames.helpers.series_helper import (
    assert_and_find_core_series,
    add_primitive_methods, has_window_function,
)
from pylegend.core.tds.pandas_api.frames.pandas_api_applied_function_tds_frame import PandasApiAppliedFunctionTdsFrame
from pylegend.core.tds.pandas_api.frames.pandas_api_groupby_tds_frame import PandasApiGroupbyTdsFrame
from pylegend.core.tds.result_handler import ResultHandler, ToStringResultHandler
from pylegend.core.tds.sql_query_helpers import create_sub_query
from pylegend.core.tds.tds_column import TdsColumn
from pylegend.core.tds.tds_frame import FrameToPureConfig, FrameToSqlConfig
from pylegend.extensions.tds.result_handler import PandasDfReadConfig, ToPandasDfResultHandler

if TYPE_CHECKING:
    from pylegend.core.tds.pandas_api.frames.pandas_api_tds_frame import PandasApiTdsFrame

__all__: PyLegendSequence[str] = [
    "GroupbySeries",
    "BooleanGroupbySeries",
    "StringGroupbySeries",
    "NumberGroupbySeries",
    "IntegerGroupbySeries",
    "FloatGroupbySeries",
    "DateGroupbySeries",
    "DateTimeGroupbySeries",
    "StrictDateGroupbySeries",
]

R = PyLegendTypeVar('R')


class GroupbySeries(PyLegendColumnExpression, PyLegendPrimitive, BaseTdsFrame):
    _base_groupby_frame: PandasApiGroupbyTdsFrame
    _applied_function_frame: PyLegendOptional[PandasApiAppliedFunctionTdsFrame]

    def __init__(
            self,
            base_groupby_frame: PandasApiGroupbyTdsFrame,
            applied_function_frame: PyLegendOptional[PandasApiAppliedFunctionTdsFrame] = None,
            expr: PyLegendOptional[PyLegendExpression] = None
    ) -> None:
        selected_columns = base_groupby_frame.get_selected_columns()
        assert selected_columns is not None and len(selected_columns) == 1, (
            "To initialize a GroupbySeries object, exactly one column must be selected, "
            f"but got selected columns: {[str(col) for col in selected_columns] if selected_columns is not None else None}"
        )

        row = PandasApiTdsRow.from_tds_frame("c", base_groupby_frame.base_frame())
        PyLegendColumnExpression.__init__(self, row=row, column=selected_columns[0].get_name())

        self._base_groupby_frame: PandasApiGroupbyTdsFrame = base_groupby_frame
        self._applied_function_frame = applied_function_frame

        self._expr = expr
        if self._expr is not None:
            assert_and_find_core_series(self._expr)

    @property
    def expr(self) -> PyLegendOptional[PyLegendExpression]:
        return self._expr

    @property
    def applied_function_frame(self) -> PyLegendOptional[PandasApiAppliedFunctionTdsFrame]:
        return self._applied_function_frame

    def raise_exception_if_no_function_applied(self) -> PandasApiAppliedFunctionTdsFrame:
        if self._applied_function_frame is None:
            raise RuntimeError(
                "The 'groupby' function requires at least one operation to be performed right after it (e.g. aggregate, rank)"
            )
        return self._applied_function_frame

    def get_base_frame(self) -> "PandasApiGroupbyTdsFrame":
        return self._base_groupby_frame

    def get_leaf_expressions(self) -> PyLegendSequence["PyLegendExpression"]:
        if self.expr is not None:
            return self.expr.get_leaf_expressions()
        return [self]

    def to_sql_expression(
            self,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        if self.expr is not None:
            return self.expr.to_sql_expression(frame_name_to_base_query_map, config)

        applied_function_frame = self.raise_exception_if_no_function_applied()
        applied_func = applied_function_frame.get_applied_function()
        if isinstance(applied_func, SupportsToSqlExpression):
            return applied_func.to_sql_expression(frame_name_to_base_query_map, config)

        raise NotImplementedError(  # pragma: no cover
            f"The '{applied_func.name()}' function cannot provide a SQL expression"
        )

    def to_pure_expression(self, config: FrameToPureConfig) -> str:
        if self._expr is not None:
            return self._expr.to_pure_expression(config)

        applied_function_frame = self.raise_exception_if_no_function_applied()
        applied_func = applied_function_frame.get_applied_function()
        if isinstance(applied_func, SupportsToPureExpression):
            return applied_func.to_pure_expression(config)

        raise NotImplementedError(  # pragma: no cover
            f"The '{applied_func.name()}' function cannot provide a pure expression"
        )

    def columns(self) -> PyLegendSequence[TdsColumn]:
        if self.has_applied_function():
            assert self.applied_function_frame is not None
            return self.applied_function_frame.columns()
        selected_columns = self.get_base_frame().get_selected_columns()
        assert selected_columns is not None and len(selected_columns) == 1
        return selected_columns

    def to_sql_query(self, config: FrameToSqlConfig = FrameToSqlConfig()) -> str:
        query = self.to_sql_query_object(config)
        sql_to_string_config = SqlToStringConfig(
            format_=SqlToStringFormat(pretty=config.pretty)
        )
        return config.sql_to_string_generator().generate_sql_string(query, sql_to_string_config)

    def to_pure_query(self, config: FrameToPureConfig = FrameToPureConfig()) -> str:
        temp_column_name_suffix = "__INTERNAL_PYLEGEND_COLUMN__"
        if self.expr is None:
            return self.raise_exception_if_no_function_applied().to_pure_query(config)

        col_name = self.columns()[0].get_name()
        full_expr = self.expr
        has_window_func = False
        window_expr = ""
        function_expr = ""
        sub_expressions = self.get_leaf_expressions()
        for expr in sub_expressions:
            if isinstance(expr, GroupbySeries):
                applied_func = expr.raise_exception_if_no_function_applied().get_applied_function()
                if isinstance(applied_func, RankFunction):
                    assert has_window_func is False
                    has_window_func = True
                    c, window = applied_func.construct_column_expression_and_window_tuples()[0]
                    window_expr = window.to_pure_expression(config)
                    function_expr = c[1].to_pure_expression(config)

        extend = ""
        if has_window_func:
            pure_expr = full_expr.to_pure_expression(config)
            temp_name = escape_column_name(col_name + temp_column_name_suffix)
            extend = f"->extend({window_expr}, ~{temp_name}:{generate_pure_lambda('p,w,r', function_expr)})"
            project = f"->project(~[{escape_column_name(col_name)}:c|{pure_expr}])"
        else:  # pragma: no cover
            project = f"->project(~[{escape_column_name(col_name)}:c|{self.to_pure_expression(config)}])"

        if len(extend) > 0:
            extend = config.separator(1) + extend
        project = config.separator(1) + project

        return self.get_base_frame().base_frame().to_pure_query(config) + extend + project

    def execute_frame(
            self,
            result_handler: ResultHandler[R],
            chunk_size: PyLegendOptional[int] = None
    ) -> R:  # pragma: no cover
        return BaseTdsFrame.execute_frame(self, result_handler, chunk_size)

    def execute_frame_to_string(
            self,
            chunk_size: PyLegendOptional[int] = None
    ) -> str:  # pragma: no cover
        return self.execute_frame(ToStringResultHandler(), chunk_size)

    def execute_frame_to_pandas_df(
            self,
            chunk_size: PyLegendOptional[int] = None,
            pandas_df_read_config: PandasDfReadConfig = PandasDfReadConfig()
    ) -> pd.DataFrame:  # pragma: no cover
        return self.execute_frame(ToPandasDfResultHandler(pandas_df_read_config), chunk_size)

    def to_sql_query_object(self, config: FrameToSqlConfig) -> QuerySpecification:
        temp_column_name_suffix = "__INTERNAL_PYLEGEND_COLUMN__"
        if self.expr is None:
            return self.raise_exception_if_no_function_applied().to_sql_query_object(config)

        expr_contains_window_func = has_window_function(self)

        db_extension = config.sql_to_string_generator().get_db_extension()
        base_query = self.get_base_frame().base_frame().to_sql_query_object(config)
        col_name = self.columns()[0].get_name()

        temp_col_name = (
            db_extension.quote_identifier(col_name + temp_column_name_suffix) if expr_contains_window_func else
            db_extension.quote_identifier(col_name)
        )

        new_select_item = SingleColumn(
            temp_col_name,
            self.to_sql_expression({'c': base_query}, config)
        )
        base_query.select.selectItems = [new_select_item]

        if expr_contains_window_func:
            new_query = create_sub_query(base_query, config, "root")
            new_query.select.selectItems = [
                SingleColumn(
                    db_extension.quote_identifier(col_name),
                    QualifiedNameReference(QualifiedName([db_extension.quote_identifier("root"), temp_col_name]))
                )
            ]
            return new_query
        else:  # pragma: no cover
            return base_query

    def to_pure(self, config: FrameToPureConfig) -> str:
        return self.to_pure_query(config)

    def get_all_tds_frames(self) -> PyLegendSequence["BaseTdsFrame"]:
        if self.expr is not None:
            core_groupby_series = assert_and_find_core_series(self)
            assert core_groupby_series is not None
            return core_groupby_series.get_all_tds_frames()
        applied_function_frame = self.raise_exception_if_no_function_applied()
        return applied_function_frame.get_all_tds_frames()

    def has_applied_function(self) -> bool:
        return self.applied_function_frame is not None

    def aggregate(
            self,
            func: PyLegendAggInput,
            axis: PyLegendUnion[int, str] = 0,
            *args: PyLegendPrimitiveOrPythonPrimitive,
            **kwargs: PyLegendPrimitiveOrPythonPrimitive
    ) -> "PandasApiTdsFrame":
        if self._expr is not None:  # pragma: no cover
            error_msg = '''
                Applying aggregate function to a computed series expression is not supported yet.
                For example,
                    not supported: (frame.groupby('grp')['col'] + 5).sum()
                    supported: frame.groupby('grp')['col'].sum() + 5
            '''
            error_msg = dedent(error_msg).strip()
            raise NotImplementedError(error_msg)

        new_series = copy.copy(self)
        if new_series.applied_function_frame is None:
            return new_series.get_base_frame().aggregate(func, axis, *args, **kwargs)
        else:
            return new_series.applied_function_frame.aggregate(func, axis, *args, **kwargs)

    def agg(
            self,
            func: PyLegendAggInput,
            axis: PyLegendUnion[int, str] = 0,
            *args: PyLegendPrimitiveOrPythonPrimitive,
            **kwargs: PyLegendPrimitiveOrPythonPrimitive
    ) -> "PandasApiTdsFrame":
        return self.aggregate(func, axis, *args, **kwargs)

    def sum(
        self,
        numeric_only: bool = False,
        min_count: int = 0,
        engine: PyLegendOptional[str] = None,
        engine_kwargs: PyLegendOptional[PyLegendDict[str, bool]] = None,
    ) -> "PandasApiTdsFrame":
        if numeric_only is not False:
            raise NotImplementedError("numeric_only=True is not currently supported in sum function.")
        if min_count != 0:
            raise NotImplementedError(f"min_count must be 0 in sum function, but got: {min_count}")
        if engine is not None:
            raise NotImplementedError("engine parameter is not supported in sum function.")
        if engine_kwargs is not None:
            raise NotImplementedError("engine_kwargs parameter is not supported in sum function.")
        return self.aggregate("sum", 0)

    def mean(
        self,
        numeric_only: bool = False,
        engine: PyLegendOptional[str] = None,
        engine_kwargs: PyLegendOptional[PyLegendDict[str, bool]] = None,
    ) -> "PandasApiTdsFrame":
        if numeric_only is not False:
            raise NotImplementedError("numeric_only=True is not currently supported in mean function.")
        if engine is not None:
            raise NotImplementedError("engine parameter is not supported in mean function.")
        if engine_kwargs is not None:
            raise NotImplementedError("engine_kwargs parameter is not supported in mean function.")
        return self.aggregate("mean", 0)

    def min(
        self,
        numeric_only: bool = False,
        min_count: int = -1,
        engine: PyLegendOptional[str] = None,
        engine_kwargs: PyLegendOptional[PyLegendDict[str, bool]] = None,
    ) -> "PandasApiTdsFrame":
        if numeric_only is not False:
            raise NotImplementedError("numeric_only=True is not currently supported in min function.")
        if min_count != -1:
            raise NotImplementedError(f"min_count must be -1 (default) in min function, but got: {min_count}")
        if engine is not None:
            raise NotImplementedError("engine parameter is not supported in min function.")
        if engine_kwargs is not None:
            raise NotImplementedError("engine_kwargs parameter is not supported in min function.")
        return self.aggregate("min", 0)

    def max(
        self,
        numeric_only: bool = False,
        min_count: int = -1,
        engine: PyLegendOptional[str] = None,
        engine_kwargs: PyLegendOptional[PyLegendDict[str, bool]] = None,
    ) -> "PandasApiTdsFrame":
        if numeric_only is not False:
            raise NotImplementedError("numeric_only=True is not currently supported in max function.")
        if min_count != -1:
            raise NotImplementedError(f"min_count must be -1 (default) in max function, but got: {min_count}")
        if engine is not None:
            raise NotImplementedError("engine parameter is not supported in max function.")
        if engine_kwargs is not None:
            raise NotImplementedError("engine_kwargs parameter is not supported in max function.")
        return self.aggregate("max", 0)

    def std(
        self,
        ddof: int = 1,
        engine: PyLegendOptional[str] = None,
        engine_kwargs: PyLegendOptional[PyLegendDict[str, bool]] = None,
        numeric_only: bool = False,
    ) -> "PandasApiTdsFrame":
        if ddof != 1:
            raise NotImplementedError(f"Only ddof=1 (Sample Standard Deviation) is supported in std function, but got: {ddof}")
        if engine is not None:
            raise NotImplementedError("engine parameter is not supported in std function.")
        if engine_kwargs is not None:
            raise NotImplementedError("engine_kwargs parameter is not supported in std function.")
        if numeric_only is not False:
            raise NotImplementedError("numeric_only=True is not currently supported in std function.")
        return self.aggregate("std", 0)

    def var(
        self,
        ddof: int = 1,
        engine: PyLegendOptional[str] = None,
        engine_kwargs: PyLegendOptional[PyLegendDict[str, bool]] = None,
        numeric_only: bool = False,
    ) -> "PandasApiTdsFrame":
        if ddof != 1:
            raise NotImplementedError(f"Only ddof=1 (Sample Variance) is supported in var function, but got: {ddof}")
        if engine is not None:
            raise NotImplementedError("engine parameter is not supported in var function.")
        if engine_kwargs is not None:
            raise NotImplementedError("engine_kwargs parameter is not supported in var function.")
        if numeric_only is not False:
            raise NotImplementedError("numeric_only=True is not currently supported in var function.")
        return self.aggregate("var", 0)

    def count(self) -> "PandasApiTdsFrame":
        return self.aggregate("count", 0)

    def rank(
            self,
            method: str = 'min',
            ascending: bool = True,
            na_option: str = 'bottom',
            pct: bool = False,
            axis: PyLegendUnion[int, str] = 0
    ) -> "GroupbySeries":
        if self._expr is not None:  # pragma: no cover
            error_msg = '''
                Applying rank function to a computed series expression is not supported yet.
                For example,
                    not supported: (frame.groupby('grp')['col'] + 5).rank()
                    supported: frame.groupby('grp')['col'].rank() + 5
            '''
            error_msg = dedent(error_msg).strip()
            raise NotImplementedError(error_msg)

        applied_function_frame = self._base_groupby_frame.rank(method, ascending, na_option, pct, axis)
        assert isinstance(applied_function_frame, PandasApiAppliedFunctionTdsFrame)

        if pct:
            return FloatGroupbySeries(self._base_groupby_frame, applied_function_frame)
        else:
            return IntegerGroupbySeries(self._base_groupby_frame, applied_function_frame)


@add_primitive_methods
class BooleanGroupbySeries(GroupbySeries, PyLegendBoolean, PyLegendExpressionBooleanReturn):
    def __init__(  # pragma: no cover (Boolean column not supported in PURE)
            self,
            base_groupby_frame: PandasApiGroupbyTdsFrame,
            applied_function_frame: PyLegendOptional[PandasApiAppliedFunctionTdsFrame] = None,
            expr: PyLegendOptional[PyLegendExpression] = None
    ) -> None:
        super().__init__(base_groupby_frame, applied_function_frame, expr)
        PyLegendBoolean.__init__(self, self)


@add_primitive_methods
class StringGroupbySeries(GroupbySeries, PyLegendString, PyLegendExpressionStringReturn):
    def __init__(
            self,
            base_groupby_frame: PandasApiGroupbyTdsFrame,
            applied_function_frame: PyLegendOptional[PandasApiAppliedFunctionTdsFrame] = None,
            expr: PyLegendOptional[PyLegendExpression] = None
    ) -> None:
        super().__init__(base_groupby_frame, applied_function_frame, expr)
        PyLegendString.__init__(self, self)


@add_primitive_methods
class NumberGroupbySeries(GroupbySeries, PyLegendNumber, PyLegendExpressionNumberReturn):
    def __init__(
            self,
            base_groupby_frame: PandasApiGroupbyTdsFrame,
            applied_function_frame: PyLegendOptional[PandasApiAppliedFunctionTdsFrame] = None,
            expr: PyLegendOptional[PyLegendExpression] = None
    ) -> None:
        super().__init__(base_groupby_frame, applied_function_frame, expr)
        PyLegendNumber.__init__(self, self)


@add_primitive_methods
class IntegerGroupbySeries(NumberGroupbySeries, PyLegendInteger, PyLegendExpressionIntegerReturn):
    def __init__(
            self,
            base_groupby_frame: PandasApiGroupbyTdsFrame,
            applied_function_frame: PyLegendOptional[PandasApiAppliedFunctionTdsFrame] = None,
            expr: PyLegendOptional[PyLegendExpression] = None
    ) -> None:
        super().__init__(base_groupby_frame, applied_function_frame, expr)
        PyLegendInteger.__init__(self, self)


@add_primitive_methods
class FloatGroupbySeries(NumberGroupbySeries, PyLegendFloat, PyLegendExpressionFloatReturn):
    def __init__(
            self,
            base_groupby_frame: PandasApiGroupbyTdsFrame,
            applied_function_frame: PyLegendOptional[PandasApiAppliedFunctionTdsFrame] = None,
            expr: PyLegendOptional[PyLegendExpression] = None
    ) -> None:
        super().__init__(base_groupby_frame, applied_function_frame, expr)
        PyLegendFloat.__init__(self, self)


@add_primitive_methods
class DecimalGroupbySeries(NumberGroupbySeries, PyLegendDecimal, PyLegendExpressionDecimalReturn):
    def __init__(
            self,
            base_groupby_frame: PandasApiGroupbyTdsFrame,
            applied_function_frame: PyLegendOptional[PandasApiAppliedFunctionTdsFrame] = None,
            expr: PyLegendOptional[PyLegendExpression] = None
    ) -> None:
        super().__init__(base_groupby_frame, applied_function_frame, expr)
        PyLegendDecimal.__init__(self, self)


@add_primitive_methods
class DateGroupbySeries(GroupbySeries, PyLegendDate, PyLegendExpressionDateReturn):
    def __init__(
            self,
            base_groupby_frame: PandasApiGroupbyTdsFrame,
            applied_function_frame: PyLegendOptional[PandasApiAppliedFunctionTdsFrame] = None,
            expr: PyLegendOptional[PyLegendExpression] = None
    ) -> None:
        super().__init__(base_groupby_frame, applied_function_frame, expr)
        PyLegendDate.__init__(self, self)


@add_primitive_methods
class DateTimeGroupbySeries(DateGroupbySeries, PyLegendDateTime, PyLegendExpressionDateTimeReturn):
    def __init__(
            self,
            base_groupby_frame: PandasApiGroupbyTdsFrame,
            applied_function_frame: PyLegendOptional[PandasApiAppliedFunctionTdsFrame] = None,
            expr: PyLegendOptional[PyLegendExpression] = None
    ) -> None:
        super().__init__(base_groupby_frame, applied_function_frame, expr)
        PyLegendDateTime.__init__(self, self)


@add_primitive_methods
class StrictDateGroupbySeries(DateGroupbySeries, PyLegendStrictDate, PyLegendExpressionStrictDateReturn):
    def __init__(
            self,
            base_groupby_frame: PandasApiGroupbyTdsFrame,
            applied_function_frame: PyLegendOptional[PandasApiAppliedFunctionTdsFrame] = None,
            expr: PyLegendOptional[PyLegendExpression] = None
    ) -> None:
        super().__init__(base_groupby_frame, applied_function_frame, expr)
        PyLegendStrictDate.__init__(self, self)



# Copyright 2025 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from textwrap import dedent
from typing import TYPE_CHECKING, runtime_checkable, Protocol

import pandas as pd

from pylegend._typing import (
    PyLegendDict,
)
from pylegend._typing import (
    PyLegendSequence,
    PyLegendOptional,
    PyLegendTypeVar,
    PyLegendUnion
)
from pylegend.core.database.sql_to_string import SqlToStringConfig, SqlToStringFormat
from pylegend.core.language.pandas_api.pandas_api_aggregate_specification import PyLegendAggInput
from pylegend.core.language.pandas_api.pandas_api_tds_row import PandasApiTdsRow
from pylegend.core.language.shared.column_expressions import PyLegendColumnExpression
from pylegend.core.language.shared.expression import (
    PyLegendExpressionBooleanReturn,
    PyLegendExpressionStringReturn,
    PyLegendExpressionNumberReturn,
    PyLegendExpressionIntegerReturn,
    PyLegendExpressionFloatReturn,
    PyLegendExpressionDecimalReturn,
    PyLegendExpressionDateReturn,
    PyLegendExpressionDateTimeReturn,
    PyLegendExpressionStrictDateReturn,
    PyLegendExpression,
)
from pylegend.core.language.shared.helpers import escape_column_name, generate_pure_lambda
from pylegend.core.language.shared.primitives.boolean import PyLegendBoolean
from pylegend.core.language.shared.primitives.date import PyLegendDate
from pylegend.core.language.shared.primitives.datetime import PyLegendDateTime
from pylegend.core.language.shared.primitives.float import PyLegendFloat
from pylegend.core.language.shared.primitives.decimal import PyLegendDecimal
from pylegend.core.language.shared.primitives.integer import PyLegendInteger
from pylegend.core.language.shared.primitives.number import PyLegendNumber
from pylegend.core.language.shared.primitives.primitive import PyLegendPrimitive, PyLegendPrimitiveOrPythonPrimitive
from pylegend.core.language.shared.primitives.strictdate import PyLegendStrictDate
from pylegend.core.language.shared.primitives.string import PyLegendString
from pylegend.core.sql.metamodel import (
    Expression, SingleColumn, QualifiedNameReference, QualifiedName,
)
from pylegend.core.sql.metamodel import QuerySpecification
from pylegend.core.tds.abstract.frames.base_tds_frame import BaseTdsFrame
from pylegend.core.tds.pandas_api.frames.functions.filter import PandasApiFilterFunction
from pylegend.core.tds.pandas_api.frames.functions.rank_function import RankFunction
from pylegend.core.tds.pandas_api.frames.helpers.series_helper import add_primitive_methods, assert_and_find_core_series, \
    has_window_function
from pylegend.core.tds.pandas_api.frames.pandas_api_applied_function_tds_frame import PandasApiAppliedFunctionTdsFrame
from pylegend.core.tds.pandas_api.frames.pandas_api_base_tds_frame import PandasApiBaseTdsFrame
from pylegend.core.tds.result_handler import ResultHandler, ToStringResultHandler
from pylegend.core.tds.sql_query_helpers import create_sub_query
from pylegend.core.tds.tds_column import TdsColumn
from pylegend.core.tds.tds_frame import FrameToPureConfig
from pylegend.core.tds.tds_frame import FrameToSqlConfig
from pylegend.extensions.tds.result_handler import PandasDfReadConfig, ToPandasDfResultHandler

if TYPE_CHECKING:
    from pylegend.core.tds.pandas_api.frames.pandas_api_tds_frame import PandasApiTdsFrame

__all__: PyLegendSequence[str] = [
    "Series",
    "BooleanSeries",
    "StringSeries",
    "NumberSeries",
    "IntegerSeries",
    "FloatSeries",
    "DateSeries",
    "DateTimeSeries",
    "StrictDateSeries",
    "SupportsToSqlExpression",
    "SupportsToPureExpression",
]

R = PyLegendTypeVar('R')


@runtime_checkable
class SupportsToSqlExpression(Protocol):
    def to_sql_expression(
            self,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        ...


@runtime_checkable
class SupportsToPureExpression(Protocol):
    def to_pure_expression(self, config: FrameToPureConfig) -> str:
        ...


@add_primitive_methods
class Series(PyLegendColumnExpression, PyLegendPrimitive, BaseTdsFrame):
    def __init__(
            self, base_frame: "PandasApiBaseTdsFrame", column: str, expr: PyLegendOptional[PyLegendExpression] = None
    ) -> None:
        row = PandasApiTdsRow.from_tds_frame("c", base_frame)
        PyLegendColumnExpression.__init__(self, row=row, column=column)

        self._base_frame = base_frame
        filtered = base_frame.filter(items=[column])
        assert isinstance(filtered, PandasApiAppliedFunctionTdsFrame)
        self._filtered_frame: PandasApiAppliedFunctionTdsFrame = filtered

        self._expr = expr
        if self._expr is not None:
            assert_and_find_core_series(self._expr)

    @property
    def expr(self) -> PyLegendOptional[PyLegendExpression]:
        return self._expr

    def value(self) -> PyLegendColumnExpression:
        return self

    def get_base_frame(self) -> "PandasApiBaseTdsFrame":
        return self._base_frame

    def get_filtered_frame(self) -> PandasApiAppliedFunctionTdsFrame:
        return self._filtered_frame

    def get_leaf_expressions(self) -> PyLegendSequence["PyLegendExpression"]:
        if self.expr is not None:
            return self.expr.get_leaf_expressions()
        return [self]

    def to_sql_expression(
            self,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        if self._expr is not None:
            return self._expr.to_sql_expression(frame_name_to_base_query_map, config)

        applied_func = self._filtered_frame.get_applied_function()
        if not isinstance(applied_func, PandasApiFilterFunction):  # pragma: no cover
            if isinstance(applied_func, SupportsToSqlExpression):
                return applied_func.to_sql_expression(frame_name_to_base_query_map, config)
            else:
                raise NotImplementedError(
                    f"The '{applied_func.name()}' function cannot provide a SQL expression"
                )

        return super().to_sql_expression(frame_name_to_base_query_map, config)

    def to_pure_expression(self, config: FrameToPureConfig) -> str:
        if self._expr is not None:
            return self._expr.to_pure_expression(config)

        applied_func = self._filtered_frame.get_applied_function()
        if not isinstance(applied_func, PandasApiFilterFunction):  # pragma: no cover
            if isinstance(applied_func, SupportsToPureExpression):
                return applied_func.to_pure_expression(config)
            else:
                raise NotImplementedError(
                    f"The '{applied_func.name()}' function cannot provide a pure expression"
                )

        return super().to_pure_expression(config)

    def columns(self) -> PyLegendSequence[TdsColumn]:
        return self._filtered_frame.columns()

    def to_sql_query(self, config: FrameToSqlConfig = FrameToSqlConfig()) -> str:
        query = self.to_sql_query_object(config)
        sql_to_string_config = SqlToStringConfig(
            format_=SqlToStringFormat(pretty=config.pretty)
        )
        return config.sql_to_string_generator().generate_sql_string(query, sql_to_string_config)

    def to_pure_query(self, config: FrameToPureConfig = FrameToPureConfig()) -> str:
        temp_column_name_suffix = "__INTERNAL_PYLEGEND_COLUMN__"
        if self.expr is None:
            return self.get_filtered_frame().to_pure_query(config)

        col_name = self.columns()[0].get_name()
        full_expr = self.expr
        has_window_func = False
        window_expr = ""
        function_expr = ""
        sub_expressions = self.get_leaf_expressions()
        for expr in sub_expressions:
            if isinstance(expr, Series):
                applied_func = expr.get_filtered_frame().get_applied_function()
                if isinstance(applied_func, RankFunction):
                    assert has_window_func is False
                    has_window_func = True
                    c, window = applied_func.construct_column_expression_and_window_tuples()[0]
                    window_expr = window.to_pure_expression(config)
                    function_expr = c[1].to_pure_expression(config)

        extend = ""
        if has_window_func:
            pure_expr = full_expr.to_pure_expression(config)
            temp_name = escape_column_name(col_name + temp_column_name_suffix)
            extend = f"->extend({window_expr}, ~{temp_name}:{generate_pure_lambda('p,w,r', function_expr)})"
            project = f"->project(~[{escape_column_name(col_name)}:c|{pure_expr}])"
        else:
            project = f"->project(~[{escape_column_name(col_name)}:c|{self.to_pure_expression(config)}])"

        if len(extend) > 0:
            extend = config.separator(1) + extend
        project = config.separator(1) + project

        return self.get_base_frame().to_pure_query(config) + extend + project

    def execute_frame(
            self,
            result_handler: ResultHandler[R],
            chunk_size: PyLegendOptional[int] = None
    ) -> R:
        return BaseTdsFrame.execute_frame(self, result_handler, chunk_size)

    def execute_frame_to_string(
            self,
            chunk_size: PyLegendOptional[int] = None
    ) -> str:
        return self.execute_frame(ToStringResultHandler(), chunk_size)

    def execute_frame_to_pandas_df(
            self,
            chunk_size: PyLegendOptional[int] = None,
            pandas_df_read_config: PandasDfReadConfig = PandasDfReadConfig()
    ) -> pd.DataFrame:
        return self.execute_frame(ToPandasDfResultHandler(pandas_df_read_config), chunk_size)  # pragma: no cover

    def to_sql_query_object(self, config: FrameToSqlConfig) -> QuerySpecification:
        temp_column_name_suffix = "__INTERNAL_PYLEGEND_COLUMN__"
        if self.expr is None:
            return self.get_filtered_frame().to_sql_query_object(config)

        expr_contains_window_func = has_window_function(self)

        db_extension = config.sql_to_string_generator().get_db_extension()
        base_query = self.get_base_frame().to_sql_query_object(config)
        col_name = self.columns()[0].get_name()

        temp_col_name = (
            db_extension.quote_identifier(col_name + temp_column_name_suffix) if expr_contains_window_func else
            db_extension.quote_identifier(col_name)
        )

        new_select_item = SingleColumn(
            temp_col_name,
            self.to_sql_expression({'c': base_query}, config)
        )
        base_query.select.selectItems = [new_select_item]

        if expr_contains_window_func:
            new_query = create_sub_query(base_query, config, "root")
            new_query.select.selectItems = [
                SingleColumn(
                    db_extension.quote_identifier(col_name),
                    QualifiedNameReference(QualifiedName([db_extension.quote_identifier("root"), temp_col_name]))
                )
            ]
            return new_query
        else:
            return base_query

    def to_pure(self, config: FrameToPureConfig) -> str:
        return self.to_pure_query(config)

    def get_all_tds_frames(self) -> PyLegendSequence["BaseTdsFrame"]:
        if self.expr is not None:
            core_series = assert_and_find_core_series(self)
            assert core_series is not None
            return core_series.get_all_tds_frames()
        return self._filtered_frame.get_all_tds_frames()

    def has_applied_function(self) -> bool:
        applied_func = self._filtered_frame.get_applied_function()
        return not isinstance(applied_func, PandasApiFilterFunction)

    def aggregate(
            self,
            func: PyLegendAggInput,
            axis: PyLegendUnion[int, str] = 0,
            *args: PyLegendPrimitiveOrPythonPrimitive,
            **kwargs: PyLegendPrimitiveOrPythonPrimitive
    ) -> "PandasApiTdsFrame":
        if self._expr is not None:  # pragma: no cover
            error_msg = '''
                Applying aggregate function to a computed series expression is not supported yet.
                Please change the series itself before trying to apply aggregate function.
                For example,
                    instead of: (frame['col'] + 5).sum()
                    do: frame['new_col'] = frame['col'] + 5; frame['new_col'].sum()
            '''
            error_msg = dedent(error_msg).strip()
            raise NotImplementedError(error_msg)
        return self._filtered_frame.aggregate(func, axis, *args, **kwargs)

    def agg(
            self,
            func: PyLegendAggInput,
            axis: PyLegendUnion[int, str] = 0,
            *args: PyLegendPrimitiveOrPythonPrimitive,
            **kwargs: PyLegendPrimitiveOrPythonPrimitive
    ) -> "PandasApiTdsFrame":
        return self.aggregate(func, axis, *args, **kwargs)

    def sum(
            self,
            axis: PyLegendUnion[int, str] = 0,
            skipna: bool = True,
            numeric_only: bool = False,
            min_count: int = 0,
            **kwargs: PyLegendPrimitiveOrPythonPrimitive
    ) -> "PandasApiTdsFrame":
        if axis not in [0, "index"]:
            raise NotImplementedError(f"The 'axis' parameter must be 0 or 'index' in sum function, but got: {axis}")
        if skipna is not True:
            raise NotImplementedError("skipna=False is not currently supported in sum function. "
                                      "SQL aggregation ignores nulls by default.")
        if numeric_only is not False:
            raise NotImplementedError("numeric_only=True is not currently supported in sum function.")
        if min_count != 0:
            raise NotImplementedError(f"min_count must be 0 in sum function, but got: {min_count}")
        if len(kwargs) > 0:
            raise NotImplementedError(
                f"Additional keyword arguments not supported in sum function: {list(kwargs.keys())}")
        return self.aggregate("sum", 0)

    def mean(
            self,
            axis: PyLegendUnion[int, str] = 0,
            skipna: bool = True,
            numeric_only: bool = False,
            **kwargs: PyLegendPrimitiveOrPythonPrimitive
    ) -> "PandasApiTdsFrame":
        if axis not in [0, "index"]:
            raise NotImplementedError(f"The 'axis' parameter must be 0 or 'index' in mean function, but got: {axis}")
        if skipna is not True:
            raise NotImplementedError("skipna=False is not currently supported in mean function.")
        if numeric_only is not False:
            raise NotImplementedError("numeric_only=True is not currently supported in mean function.")
        if len(kwargs) > 0:
            raise NotImplementedError(
                f"Additional keyword arguments not supported in mean function: {list(kwargs.keys())}")
        return self.aggregate("mean", 0)

    def min(
            self,
            axis: PyLegendUnion[int, str] = 0,
            skipna: bool = True,
            numeric_only: bool = False,
            **kwargs: PyLegendPrimitiveOrPythonPrimitive
    ) -> "PandasApiTdsFrame":
        if axis not in [0, "index"]:
            raise NotImplementedError(f"The 'axis' parameter must be 0 or 'index' in min function, but got: {axis}")
        if skipna is not True:
            raise NotImplementedError("skipna=False is not currently supported in min function.")
        if numeric_only is not False:
            raise NotImplementedError("numeric_only=True is not currently supported in min function.")
        if len(kwargs) > 0:
            raise NotImplementedError(
                f"Additional keyword arguments not supported in min function: {list(kwargs.keys())}")
        return self.aggregate("min", 0)

    def max(
            self,
            axis: PyLegendUnion[int, str] = 0,
            skipna: bool = True,
            numeric_only: bool = False,
            **kwargs: PyLegendPrimitiveOrPythonPrimitive
    ) -> "PandasApiTdsFrame":
        if axis not in [0, "index"]:
            raise NotImplementedError(f"The 'axis' parameter must be 0 or 'index' in max function, but got: {axis}")
        if skipna is not True:
            raise NotImplementedError("skipna=False is not currently supported in max function.")
        if numeric_only is not False:
            raise NotImplementedError("numeric_only=True is not currently supported in max function.")
        if len(kwargs) > 0:
            raise NotImplementedError(
                f"Additional keyword arguments not supported in max function: {list(kwargs.keys())}")
        return self.aggregate("max", 0)

    def std(
            self,
            axis: PyLegendUnion[int, str] = 0,
            skipna: bool = True,
            ddof: int = 1,
            numeric_only: bool = False,
            **kwargs: PyLegendPrimitiveOrPythonPrimitive
    ) -> "PandasApiTdsFrame":
        if axis not in [0, "index"]:
            raise NotImplementedError(f"The 'axis' parameter must be 0 or 'index' in std function, but got: {axis}")
        if skipna is not True:
            raise NotImplementedError("skipna=False is not currently supported in std function.")
        if ddof != 1:
            raise NotImplementedError(
                f"Only ddof=1 (Sample Standard Deviation) is supported in std function, but got: {ddof}")
        if numeric_only is not False:
            raise NotImplementedError("numeric_only=True is not currently supported in std function.")
        if len(kwargs) > 0:
            raise NotImplementedError(
                f"Additional keyword arguments not supported in std function: {list(kwargs.keys())}")
        return self.aggregate("std", 0)

    def var(
            self,
            axis: PyLegendUnion[int, str] = 0,
            skipna: bool = True,
            ddof: int = 1,
            numeric_only: bool = False,
            **kwargs: PyLegendPrimitiveOrPythonPrimitive
    ) -> "PandasApiTdsFrame":
        if axis not in [0, "index"]:
            raise NotImplementedError(f"The 'axis' parameter must be 0 or 'index' in var function, but got: {axis}")
        if skipna is not True:
            raise NotImplementedError("skipna=False is not currently supported in var function.")
        if ddof != 1:
            raise NotImplementedError(f"Only ddof=1 (Sample Variance) is supported in var function, but got: {ddof}")
        if numeric_only is not False:
            raise NotImplementedError("numeric_only=True is not currently supported in var function.")
        if len(kwargs) > 0:
            raise NotImplementedError(
                f"Additional keyword arguments not supported in var function: {list(kwargs.keys())}")
        return self.aggregate("var", 0)

    def count(
            self,
            axis: PyLegendUnion[int, str] = 0,
            numeric_only: bool = False,
            **kwargs: PyLegendPrimitiveOrPythonPrimitive
    ) -> "PandasApiTdsFrame":
        if axis not in [0, "index"]:
            raise NotImplementedError(f"The 'axis' parameter must be 0 or 'index' in count function, but got: {axis}")
        if numeric_only is not False:
            raise NotImplementedError("numeric_only=True is not currently supported in count function.")
        if len(kwargs) > 0:
            raise NotImplementedError(
                f"Additional keyword arguments not supported in count function: {list(kwargs.keys())}")
        return self.aggregate("count", 0)

    def rank(
            self,
            axis: PyLegendUnion[int, str] = 0,
            method: str = 'min',
            numeric_only: bool = False,
            na_option: str = 'bottom',
            ascending: bool = True,
            pct: bool = False
    ) -> "Series":
        if self._expr is not None:  # pragma: no cover
            error_msg = '''
                Applying rank function to a computed series expression is not supported yet.
                For example,
                    not supported: (frame['col'] + 5).rank()
                    supported: frame['col'].rank() + 5
            '''
            error_msg = dedent(error_msg).strip()
            raise NotImplementedError(error_msg)

        new_series: Series
        if pct:
            new_series = FloatSeries(self._filtered_frame, self.columns()[0].get_name())
        else:
            new_series = IntegerSeries(self._filtered_frame, self.columns()[0].get_name())
        new_series._base_frame = self._base_frame

        applied_function_frame = self._filtered_frame.rank(axis, method, numeric_only, na_option, ascending, pct)
        assert isinstance(applied_function_frame, PandasApiAppliedFunctionTdsFrame)

        new_series._filtered_frame = applied_function_frame
        return new_series


@add_primitive_methods
class BooleanSeries(Series, PyLegendBoolean, PyLegendExpressionBooleanReturn):  # type: ignore
    def __init__(
            self, base_frame: "PandasApiBaseTdsFrame", column: str, value: PyLegendOptional[PyLegendExpression] = None
    ) -> None:
        super().__init__(base_frame, column, value)  # pragma: no cover (Boolean column not supported in PURE)
        PyLegendBoolean.__init__(self, self)  # pragma: no cover (Boolean column not supported in PURE)


@add_primitive_methods
class StringSeries(Series, PyLegendString, PyLegendExpressionStringReturn):  # type: ignore
    def __init__(
            self, base_frame: "PandasApiBaseTdsFrame", column: str, value: PyLegendOptional[PyLegendExpression] = None
    ) -> None:
        super().__init__(base_frame, column, value)
        PyLegendString.__init__(self, self)


@add_primitive_methods
class NumberSeries(Series, PyLegendNumber, PyLegendExpressionNumberReturn):  # type: ignore
    def __init__(
            self, base_frame: "PandasApiBaseTdsFrame", column: str, value: PyLegendOptional[PyLegendExpression] = None
    ) -> None:
        super().__init__(base_frame, column, value)
        PyLegendNumber.__init__(self, self)


@add_primitive_methods
class IntegerSeries(NumberSeries, PyLegendInteger, PyLegendExpressionIntegerReturn):  # type: ignore
    def __init__(
            self, base_frame: "PandasApiBaseTdsFrame", column: str, value: PyLegendOptional[PyLegendExpression] = None
    ) -> None:
        super().__init__(base_frame, column, value)
        PyLegendInteger.__init__(self, self)


@add_primitive_methods
class FloatSeries(NumberSeries, PyLegendFloat, PyLegendExpressionFloatReturn):  # type: ignore
    def __init__(
            self, base_frame: "PandasApiBaseTdsFrame", column: str, value: PyLegendOptional[PyLegendExpression] = None
    ) -> None:
        super().__init__(base_frame, column, value)
        PyLegendFloat.__init__(self, self)


@add_primitive_methods
class DecimalSeries(NumberSeries, PyLegendDecimal, PyLegendExpressionDecimalReturn):  # type: ignore
    def __init__(
            self, base_frame: "PandasApiBaseTdsFrame", column: str, value: PyLegendOptional[PyLegendExpression] = None
    ) -> None:
        super().__init__(base_frame, column, value)
        PyLegendDecimal.__init__(self, self)


@add_primitive_methods
class DateSeries(Series, PyLegendDate, PyLegendExpressionDateReturn):  # type: ignore
    def __init__(
            self, base_frame: "PandasApiBaseTdsFrame", column: str, value: PyLegendOptional[PyLegendExpression] = None
    ) -> None:
        super().__init__(base_frame, column, value)
        PyLegendDate.__init__(self, self)


@add_primitive_methods
class DateTimeSeries(DateSeries, PyLegendDateTime, PyLegendExpressionDateTimeReturn):  # type: ignore
    def __init__(
            self, base_frame: "PandasApiBaseTdsFrame", column: str, value: PyLegendOptional[PyLegendExpression] = None
    ) -> None:
        super().__init__(base_frame, column, value)
        PyLegendDateTime.__init__(self, self)


@add_primitive_methods
class StrictDateSeries(DateSeries, PyLegendStrictDate, PyLegendExpressionStrictDateReturn):  # type: ignore
    def __init__(
            self, base_frame: "PandasApiBaseTdsFrame", column: str, value: PyLegendOptional[PyLegendExpression] = None
    ) -> None:
        super().__init__(base_frame, column, value)
        PyLegendStrictDate.__init__(self, self)



# Copyright 2026 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pylegend._typing import (
    PyLegendSequence,
    PyLegendDict,
)
from pylegend.core.language.shared.primitives.integer import PyLegendInteger
from pylegend.core.language.shared.primitives.string import PyLegendString
from pylegend.core.language.shared.primitives.float import PyLegendFloat
from pylegend.core.language.shared.primitives.datetime import PyLegendDateTime
from pylegend.core.language.shared.primitives.decimal import PyLegendDecimal
from pylegend.core.language.shared.expression import (
    PyLegendExpressionIntegerReturn,
    PyLegendExpressionStringReturn,
    PyLegendExpressionFloatReturn,
    PyLegendExpressionDateTimeReturn,
    PyLegendExpressionDecimalReturn,
)
from pylegend.core.sql.metamodel import (
    Expression,
    QuerySpecification,
)
from pylegend.core.tds.tds_frame import FrameToSqlConfig


__all__: PyLegendSequence[str] = [
    "PyLegendTinyInt",
    "PyLegendUTinyInt",
    "PyLegendSmallInt",
    "PyLegendUSmallInt",
    "PyLegendInt",
    "PyLegendUInt",
    "PyLegendBigInt",
    "PyLegendUBigInt",
    "PyLegendVarchar",
    "PyLegendTimestamp",
    "PyLegendFloat4",
    "PyLegendDouble",
    "PyLegendNumeric",
]


class PyLegendTinyInt(PyLegendInteger):
    """Precise primitive: TinyInt – signed 8-bit integer (-128 .. 127)."""

    def __init__(self, value: PyLegendExpressionIntegerReturn) -> None:
        super().__init__(value)

    def to_sql_expression(
            self,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return super().to_sql_expression(frame_name_to_base_query_map, config)


class PyLegendUTinyInt(PyLegendInteger):
    """Precise primitive: UTinyInt – unsigned 8-bit integer (0 .. 255)."""

    def __init__(self, value: PyLegendExpressionIntegerReturn) -> None:
        super().__init__(value)

    def to_sql_expression(
            self,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return super().to_sql_expression(frame_name_to_base_query_map, config)


class PyLegendSmallInt(PyLegendInteger):
    """Precise primitive: SmallInt – signed 16-bit integer (-32768 .. 32767)."""

    def __init__(self, value: PyLegendExpressionIntegerReturn) -> None:
        super().__init__(value)

    def to_sql_expression(
            self,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return super().to_sql_expression(frame_name_to_base_query_map, config)


class PyLegendUSmallInt(PyLegendInteger):
    """Precise primitive: USmallInt – unsigned 16-bit integer (0 .. 65535)."""

    def __init__(self, value: PyLegendExpressionIntegerReturn) -> None:
        super().__init__(value)

    def to_sql_expression(
            self,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return super().to_sql_expression(frame_name_to_base_query_map, config)


class PyLegendInt(PyLegendInteger):
    """Precise primitive: Int – signed 32-bit integer."""

    def __init__(self, value: PyLegendExpressionIntegerReturn) -> None:
        super().__init__(value)

    def to_sql_expression(
            self,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return super().to_sql_expression(frame_name_to_base_query_map, config)


class PyLegendUInt(PyLegendInteger):
    """Precise primitive: UInt – unsigned 32-bit integer."""

    def __init__(self, value: PyLegendExpressionIntegerReturn) -> None:
        super().__init__(value)

    def to_sql_expression(
            self,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return super().to_sql_expression(frame_name_to_base_query_map, config)


class PyLegendBigInt(PyLegendInteger):
    """Precise primitive: BigInt – signed 64-bit integer."""

    def __init__(self, value: PyLegendExpressionIntegerReturn) -> None:
        super().__init__(value)

    def to_sql_expression(
            self,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return super().to_sql_expression(frame_name_to_base_query_map, config)


class PyLegendUBigInt(PyLegendInteger):
    """Precise primitive: UBigInt – unsigned 64-bit integer."""

    def __init__(self, value: PyLegendExpressionIntegerReturn) -> None:
        super().__init__(value)

    def to_sql_expression(
            self,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return super().to_sql_expression(frame_name_to_base_query_map, config)


class PyLegendVarchar(PyLegendString):
    """Precise primitive: Varchar(max_length) – variable-length string with max length constraint."""
    __max_length: int

    def __init__(self, value: PyLegendExpressionStringReturn, max_length: int) -> None:
        super().__init__(value)
        self.__max_length = max_length

    @property
    def max_length(self) -> int:
        return self.__max_length  # pragma: no cover

    def to_sql_expression(
            self,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return super().to_sql_expression(frame_name_to_base_query_map, config)


class PyLegendTimestamp(PyLegendDateTime):
    """Precise primitive: Timestamp – extends DateTime."""

    def __init__(self, value: PyLegendExpressionDateTimeReturn) -> None:
        super().__init__(value)

    def to_sql_expression(
            self,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return super().to_sql_expression(frame_name_to_base_query_map, config)


class PyLegendFloat4(PyLegendFloat):
    """Precise primitive: Float4 – single-precision float."""

    def __init__(self, value: PyLegendExpressionFloatReturn) -> None:
        super().__init__(value)

    def to_sql_expression(
            self,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return super().to_sql_expression(frame_name_to_base_query_map, config)


class PyLegendDouble(PyLegendFloat):
    """Precise primitive: Double – double-precision float."""

    def __init__(self, value: PyLegendExpressionFloatReturn) -> None:
        super().__init__(value)

    def to_sql_expression(
            self,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return super().to_sql_expression(frame_name_to_base_query_map, config)


class PyLegendNumeric(PyLegendDecimal):
    """Precise primitive: Numeric(precision, scale) – fixed-point decimal with precision and scale constraints."""
    __precision: int
    __scale: int

    def __init__(self, value: PyLegendExpressionDecimalReturn, precision: int, scale: int) -> None:
        super().__init__(value)
        if precision < scale:
            raise ValueError(  # pragma: no cover
                f"Numeric precision ({precision}) must be >= scale ({scale})"
            )
        self.__precision = precision
        self.__scale = scale

    @property
    def precision(self) -> int:
        return self.__precision  # pragma: no cover

    @property
    def scale(self) -> int:
        return self.__scale  # pragma: no cover

    def to_sql_expression(
            self,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return super().to_sql_expression(frame_name_to_base_query_map, config)



# Copyright 2026 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pylegend._typing import (
    PyLegendSequence,
    PyLegendDict,
    PyLegendList,
    PyLegendSet,
    PyLegendUnion,
    PyLegendTuple,
    PyLegendType,
)
from pylegend.core.tds.tds_column import TdsColumn, PrimitiveTdsColumn, PrimitiveType
from pylegend.core.language.shared.primitives.primitive import PyLegendPrimitive
from pylegend.core.language.shared.primitives.boolean import PyLegendBoolean
from pylegend.core.language.shared.primitives.string import PyLegendString
from pylegend.core.language.shared.primitives.number import PyLegendNumber
from pylegend.core.language.shared.primitives.integer import PyLegendInteger
from pylegend.core.language.shared.primitives.float import PyLegendFloat
from pylegend.core.language.shared.primitives.decimal import PyLegendDecimal
from pylegend.core.language.shared.primitives.date import PyLegendDate
from pylegend.core.language.shared.primitives.datetime import PyLegendDateTime
from pylegend.core.language.shared.primitives.strictdate import PyLegendStrictDate
from pylegend.core.language.shared.primitives.precise_primitives import (
    PyLegendTinyInt,
    PyLegendUTinyInt,
    PyLegendSmallInt,
    PyLegendUSmallInt,
    PyLegendInt,
    PyLegendUInt,
    PyLegendBigInt,
    PyLegendUBigInt,
    PyLegendVarchar,
    PyLegendTimestamp,
    PyLegendFloat4,
    PyLegendDouble,
    PyLegendNumeric,
)

# Type that users pass as a cast target value:
#   - PrimitiveType.Integer                   (simple)
#   - (PrimitiveType.Numeric, 10, 2)          (with precision, scale)
#   - (PrimitiveType.Varchar, 200)            (with max length)
CastTarget = PyLegendUnion[PrimitiveType, PyLegendTuple[PrimitiveType, ...]]

__all__: PyLegendSequence[str] = [
    "CastTarget",
    "PRIMITIVE_TYPE_TO_PYLEGEND_CLASS",
    "is_cast_allowed",
    "validate_and_build_cast_columns",
    "pure_type_spec",
]

# Map each PrimitiveType enum to its corresponding PyLegend class.
# Cast validity is determined by whether the source and target classes
# have a subclass relationship (one is a subclass of the other).
PRIMITIVE_TYPE_TO_PYLEGEND_CLASS: PyLegendDict[PrimitiveType, PyLegendType[PyLegendPrimitive]] = {
    PrimitiveType.Boolean:    PyLegendBoolean,
    PrimitiveType.String:     PyLegendString,
    PrimitiveType.Number:     PyLegendNumber,
    PrimitiveType.Integer:    PyLegendInteger,
    PrimitiveType.Float:      PyLegendFloat,
    PrimitiveType.Decimal:    PyLegendDecimal,
    PrimitiveType.Date:       PyLegendDate,
    PrimitiveType.DateTime:   PyLegendDateTime,
    PrimitiveType.StrictDate: PyLegendStrictDate,
    PrimitiveType.TinyInt:    PyLegendTinyInt,
    PrimitiveType.UTinyInt:   PyLegendUTinyInt,
    PrimitiveType.SmallInt:   PyLegendSmallInt,
    PrimitiveType.USmallInt:  PyLegendUSmallInt,
    PrimitiveType.Int:        PyLegendInt,
    PrimitiveType.UInt:       PyLegendUInt,
    PrimitiveType.BigInt:     PyLegendBigInt,
    PrimitiveType.UBigInt:    PyLegendUBigInt,
    PrimitiveType.Varchar:    PyLegendVarchar,
    PrimitiveType.Timestamp:  PyLegendTimestamp,
    PrimitiveType.Float4:     PyLegendFloat4,
    PrimitiveType.Double:     PyLegendDouble,
    PrimitiveType.Numeric:    PyLegendNumeric,
}

# Types that require parameters in Pure syntax.
_PARAMETERIZED_TYPES: PyLegendSet[PrimitiveType] = {
    PrimitiveType.Numeric,
    PrimitiveType.Varchar,
}


def _normalize_target(target: CastTarget) -> PyLegendTuple[PrimitiveType, PyLegendTuple[int, ...]]:
    """Extract (PrimitiveType, params) from a CastTarget."""
    if isinstance(target, PrimitiveType):
        return target, ()
    ptype = target[0]
    params = tuple(target[1:])
    return ptype, params  # type: ignore


def pure_type_spec(ptype: PrimitiveType, params: PyLegendTuple[int, ...] = ()) -> str:
    """Return the Pure type string, e.g. 'Integer', 'Numeric(10, 2)', 'Varchar(200)'."""
    if params:
        param_str = ", ".join(str(p) for p in params)
        return f"{ptype.name}({param_str})"
    return ptype.name


def is_cast_allowed(source: PrimitiveType, target: PrimitiveType) -> bool:
    """A cast is allowed if the source and target PyLegend classes have
    a subclass relationship (one is a subclass of the other)."""
    source_cls = PRIMITIVE_TYPE_TO_PYLEGEND_CLASS.get(source)
    target_cls = PRIMITIVE_TYPE_TO_PYLEGEND_CLASS.get(target)
    if source_cls is None or target_cls is None:
        return False  # pragma: no cover
    return issubclass(source_cls, target_cls) or issubclass(target_cls, source_cls)


def validate_and_build_cast_columns(
        current_columns: PyLegendSequence[TdsColumn],
        column_type_map: PyLegendDict[str, CastTarget],
) -> PyLegendList[TdsColumn]:
    """Validate the cast request and return the new column list.

    Raises ValueError / TypeError on invalid requests.
    """
    current_col_names = {c.get_name() for c in current_columns}

    unknown_cols = set(column_type_map.keys()) - current_col_names
    if unknown_cols:
        raise ValueError(
            f"Column(s) not found in frame: {sorted(unknown_cols)}. "
            f"Available columns: {sorted(current_col_names)}"
        )

    new_columns: PyLegendList[TdsColumn] = []
    for col in current_columns:
        if col.get_name() in column_type_map:
            if not isinstance(col, PrimitiveTdsColumn):
                raise TypeError(
                    f"Cannot cast non-primitive column '{col.get_name()}' "
                    f"(type: {col.get_type()}). Only PrimitiveTdsColumn can be cast."
                )
            source_type = PrimitiveType[col.get_type()]
            target_type, params = _normalize_target(column_type_map[col.get_name()])
            if target_type in _PARAMETERIZED_TYPES and not params:
                raise ValueError(  # pragma: no cover
                    f"Cast to {target_type.name} requires parameters. "
                    f"Use a tuple, e.g. (PrimitiveType.{target_type.name}, ...) "
                    f"instead of PrimitiveType.{target_type.name}"
                )
            if not is_cast_allowed(source_type, target_type):
                source_cls = PRIMITIVE_TYPE_TO_PYLEGEND_CLASS.get(source_type)
                target_cls = PRIMITIVE_TYPE_TO_PYLEGEND_CLASS.get(target_type)
                raise ValueError(
                    f"Cannot cast column '{col.get_name()}' from {source_type.name} to {target_type.name}. "
                    f"{source_cls.__name__ if source_cls else source_type.name} and "
                    f"{target_cls.__name__ if target_cls else target_type.name} "
                    f"do not share a subclass relationship."
                )
            new_columns.append(
                PrimitiveTdsColumn(col.get_name(), target_type)
            )
        else:
            new_columns.append(col.copy())
    return new_columns



# Copyright 2025 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABCMeta
from enum import Enum
from pylegend.core.language import (
    PyLegendPrimitive,
    PyLegendBoolean,
    PyLegendString,
    PyLegendNumber,
    PyLegendInteger,
    PyLegendFloat,
    PyLegendDate,
    PyLegendDateTime,
    PyLegendStrictDate,
)
from pylegend._typing import (
    PyLegendSequence,
    PyLegendOptional,
    PyLegendList,
    PyLegendDict,
    TYPE_CHECKING,
)
from pylegend.core.language.shared.expression import (
    PyLegendExpressionFloatReturn,
    PyLegendExpressionIntegerReturn,
)
from pylegend.core.language.shared.helpers import escape_column_name
from pylegend.core.sql.metamodel import (
    Expression,
    FunctionCall,
    QualifiedName,
    QuerySpecification,
    SingleColumn,
    SortItem,
    SortItemNullOrdering,
    SortItemOrdering,
    Window
)
from pylegend.core.tds.tds_frame import (
    FrameToPureConfig,
    FrameToSqlConfig,
)

__all__: PyLegendSequence[str] = [
    "PandasApiPrimitive",
    "PandasApiBoolean",
    "PandasApiString",
    "PandasApiNumber",
    "PandasApiInteger",
    "PandasApiFloat",
    "PandasApiDate",
    "PandasApiDateTime",
    "PandasApiStrictDate",
    "PandasApiSortInfo",
    "PandasApiSortDirection",
    "PandasApiWindow",
    "PandasApiWindowReference",
    "PandasApiWindowFrame",
    "PandasApiRankExpression",
    "PandasApiDenseRankExpression",
    "PandasApiRowNumberExpression",
    "PandasApiPartialFrame",
    "PandasApiPercentRankExpression",
]


class PandasApiPrimitive(PyLegendPrimitive, metaclass=ABCMeta):
    pass


class PandasApiBoolean(PandasApiPrimitive, PyLegendBoolean):
    def __init__(self, expr: PyLegendBoolean):
        PyLegendBoolean.__init__(self, expr.value())


class PandasApiString(PandasApiPrimitive, PyLegendString):
    def __init__(self, expr: PyLegendString):
        PyLegendString.__init__(self, expr.value())


class PandasApiNumber(PandasApiPrimitive, PyLegendNumber):
    def __init__(self, expr: PyLegendNumber):
        PyLegendNumber.__init__(self, expr.value())


class PandasApiInteger(PandasApiPrimitive, PyLegendInteger):
    def __init__(self, expr: PyLegendInteger):
        PyLegendInteger.__init__(self, expr.value())


class PandasApiFloat(PandasApiPrimitive, PyLegendFloat):
    def __init__(self, expr: PyLegendFloat):
        PyLegendFloat.__init__(self, expr.value())


class PandasApiDate(PandasApiPrimitive, PyLegendDate):
    def __init__(self, expr: PyLegendDate):
        PyLegendDate.__init__(self, expr.value())


class PandasApiDateTime(PandasApiPrimitive, PyLegendDateTime):
    def __init__(self, expr: PyLegendDateTime):
        PyLegendDateTime.__init__(self, expr.value())


class PandasApiStrictDate(PandasApiPrimitive, PyLegendStrictDate):
    def __init__(self, expr: PyLegendStrictDate):
        PyLegendStrictDate.__init__(self, expr.value())


class PandasApiSortDirection(Enum):
    ASC = 1
    DESC = 2


class PandasApiSortInfo:
    __column: str
    __direction: PandasApiSortDirection
    __null_ordering: SortItemNullOrdering

    def __init__(
            self,
            column_name: str,
            direction: PandasApiSortDirection,
            null_ordering: SortItemNullOrdering = SortItemNullOrdering.UNDEFINED
    ) -> None:
        self.__column = column_name
        self.__direction = direction
        self.__null_ordering = null_ordering

    def to_sql_node(
            self,
            query: QuerySpecification,
            config: FrameToSqlConfig
    ) -> SortItem:
        return SortItem(
            sortKey=self.__find_column_expression(query, config),
            ordering=(SortItemOrdering.ASCENDING if self.__direction == PandasApiSortDirection.ASC
                      else SortItemOrdering.DESCENDING),
            nullOrdering=self.__null_ordering
        )

    def __find_column_expression(self, query: QuerySpecification, config: FrameToSqlConfig) -> Expression:
        db_extension = config.sql_to_string_generator().get_db_extension()
        filtered = [
            s for s in query.select.selectItems
            if (isinstance(s, SingleColumn) and
                s.alias == db_extension.quote_identifier(self.__column))
        ]
        if len(filtered) == 0:
            raise RuntimeError("Cannot find column: " + self.__column)  # pragma: no cover
        return filtered[0].expression

    def to_pure_expression(self, config: FrameToPureConfig) -> str:
        func = 'ascending' if self.__direction == PandasApiSortDirection.ASC else 'descending'
        return f"{func}(~{escape_column_name(self.__column)})"


class PandasApiWindowFrame(metaclass=ABCMeta):
    pass


class PandasApiWindow:
    __partition_by: PyLegendOptional[PyLegendList[str]]
    __order_by: PyLegendOptional[PyLegendList[PandasApiSortInfo]]
    __frame: PyLegendOptional[PandasApiWindowFrame]

    def __init__(
            self,
            partition_by: PyLegendOptional[PyLegendList[str]] = None,
            order_by: PyLegendOptional[PyLegendList[PandasApiSortInfo]] = None,
            frame: PyLegendOptional[PandasApiWindowFrame] = None
    ) -> None:
        self.__partition_by = partition_by
        self.__order_by = order_by
        self.__frame = frame

    def to_sql_node(
            self,
            query: QuerySpecification,
            config: FrameToSqlConfig
    ) -> Window:
        return Window(
            windowRef=None,
            partitions=(
                [] if self.__partition_by is None else
                [PandasApiWindow.__find_column_expression(query, col, config) for col in self.__partition_by]
            ),
            orderBy=(
                [] if self.__order_by is None else
                [sort_info.to_sql_node(query, config) for sort_info in self.__order_by]
            ),
            windowFrame=None
        )

    @staticmethod
    def __find_column_expression(query: QuerySpecification, col: str, config: FrameToSqlConfig) -> Expression:
        db_extension = config.sql_to_string_generator().get_db_extension()
        filtered = [
            s for s in query.select.selectItems
            if (isinstance(s, SingleColumn) and
                s.alias == db_extension.quote_identifier(col))
        ]
        if len(filtered) == 0:
            raise RuntimeError("Cannot find column: " + col)  # pragma: no cover
        return filtered[0].expression

    def to_pure_expression(self, config: FrameToPureConfig) -> str:
        partitions_str = (
            "" if self.__partition_by is None or len(self.__partition_by) == 0
            else "~[" + (', '.join(map(escape_column_name, self.__partition_by))) + "], "
        )
        sorts_str = (
            "[]" if self.__order_by is None or len(self.__order_by) == 0
            else "[" + (', '.join([s.to_pure_expression(config) for s in self.__order_by])) + "]"
        )
        return f"over({partitions_str}{sorts_str})"


class PandasApiPartialFrame:
    if TYPE_CHECKING:
        from pylegend.core.tds.pandas_api.frames.pandas_api_base_tds_frame import PandasApiBaseTdsFrame
        from pylegend.core.language.pandas_api.pandas_api_tds_row import PandasApiTdsRow

    __base_frame: "PandasApiBaseTdsFrame"
    __var_name: str

    def __init__(self, base_frame: "PandasApiBaseTdsFrame", var_name: str) -> None:
        self.__base_frame = base_frame
        self.__var_name = var_name

    def row_number(
            self,
            row: "PandasApiTdsRow"
    ) -> PyLegendInteger:
        return PyLegendInteger(PandasApiRowNumberExpression(self, row))

    def rank(
            self,
            window: "PandasApiWindowReference",
            row: "PandasApiTdsRow"
    ) -> PyLegendInteger:
        return PyLegendInteger(PandasApiRankExpression(self, window, row))

    def dense_rank(
            self,
            window: "PandasApiWindowReference",
            row: "PandasApiTdsRow"
    ) -> PyLegendInteger:
        return PyLegendInteger(PandasApiDenseRankExpression(self, window, row))

    def percent_rank(
            self,
            window: "PandasApiWindowReference",
            row: "PandasApiTdsRow"
    ) -> PyLegendFloat:
        return PyLegendFloat(PandasApiPercentRankExpression(self, window, row))

    def to_pure_expression(self, config: FrameToPureConfig) -> str:
        return f"${self.__var_name}"


class PandasApiWindowReference:
    __window: PandasApiWindow
    __var_name: str

    def __init__(self, window: PandasApiWindow, var_name: str) -> None:
        self.__window = window
        self.__var_name = var_name

    def to_pure_expression(self, config: FrameToPureConfig) -> str:
        return f"${self.__var_name}"


class PandasApiRowNumberExpression(PyLegendExpressionIntegerReturn):
    if TYPE_CHECKING:
        from pylegend.core.language.pandas_api.pandas_api_tds_row import PandasApiTdsRow

    __partial_frame: PandasApiPartialFrame
    __row: "PandasApiTdsRow"

    def __init__(
            self,
            partial_frame: PandasApiPartialFrame,
            row: "PandasApiTdsRow"
    ) -> None:
        self.__partial_frame = partial_frame
        self.__row = row

    def to_sql_expression(
            self,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return FunctionCall(
            name=QualifiedName(parts=["row_number"]), distinct=False, arguments=[], filter_=None, window=None
        )

    def to_pure_expression(self, config: FrameToPureConfig) -> str:
        return f"{self.__partial_frame.to_pure_expression(config)}->rowNumber({self.__row.to_pure_expression(config)})"


class PandasApiRankExpression(PyLegendExpressionIntegerReturn):
    if TYPE_CHECKING:
        from pylegend.core.language.pandas_api.pandas_api_tds_row import PandasApiTdsRow

    __partial_frame: PandasApiPartialFrame
    __window_ref: "PandasApiWindowReference"
    __row: "PandasApiTdsRow"

    def __init__(
            self,
            partial_frame: PandasApiPartialFrame,
            window_ref: "PandasApiWindowReference",
            row: "PandasApiTdsRow"
    ) -> None:
        self.__partial_frame = partial_frame
        self.__window_ref = window_ref
        self.__row = row

    def to_sql_expression(
            self,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return FunctionCall(
            name=QualifiedName(parts=["rank"]), distinct=False, arguments=[], filter_=None, window=None
        )

    def to_pure_expression(self, config: FrameToPureConfig) -> str:
        return (f"{self.__partial_frame.to_pure_expression(config)}->rank("
                f"{self.__window_ref.to_pure_expression(config)}, {self.__row.to_pure_expression(config)})")


class PandasApiDenseRankExpression(PyLegendExpressionIntegerReturn):
    if TYPE_CHECKING:
        from pylegend.core.language.pandas_api.pandas_api_tds_row import PandasApiTdsRow

    __partial_frame: PandasApiPartialFrame
    __window_ref: "PandasApiWindowReference"
    __row: "PandasApiTdsRow"

    def __init__(
            self,
            partial_frame: PandasApiPartialFrame,
            window_ref: "PandasApiWindowReference",
            row: "PandasApiTdsRow"
    ) -> None:
        self.__partial_frame = partial_frame
        self.__window_ref = window_ref
        self.__row = row

    def to_sql_expression(
            self,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return FunctionCall(
            name=QualifiedName(parts=["dense_rank"]), distinct=False, arguments=[], filter_=None, window=None
        )

    def to_pure_expression(self, config: FrameToPureConfig) -> str:
        return (f"{self.__partial_frame.to_pure_expression(config)}->denseRank("
                f"{self.__window_ref.to_pure_expression(config)}, {self.__row.to_pure_expression(config)})")


class PandasApiPercentRankExpression(PyLegendExpressionFloatReturn):
    if TYPE_CHECKING:
        from pylegend.core.language.pandas_api.pandas_api_tds_row import PandasApiTdsRow

    __partial_frame: PandasApiPartialFrame
    __window_ref: "PandasApiWindowReference"
    __row: "PandasApiTdsRow"

    def __init__(
            self,
            partial_frame: PandasApiPartialFrame,
            window_ref: "PandasApiWindowReference",
            row: "PandasApiTdsRow"
    ) -> None:
        self.__partial_frame = partial_frame
        self.__window_ref = window_ref
        self.__row = row

    def to_sql_expression(
            self,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return FunctionCall(
            name=QualifiedName(parts=["percent_rank"]), distinct=False, arguments=[], filter_=None, window=None
        )

    def to_pure_expression(self, config: FrameToPureConfig) -> str:
        return (f"{self.__partial_frame.to_pure_expression(config)}->percentRank("
                f"{self.__window_ref.to_pure_expression(config)}, {self.__row.to_pure_expression(config)})")



# Copyright 2026 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pylegend._typing import (
    PyLegendDict,
    PyLegendList,
    PyLegendSequence,
)
from pylegend.core.sql.metamodel import (
    QuerySpecification,
)
from pylegend.core.tds.pandas_api.frames.pandas_api_applied_function_tds_frame import PandasApiAppliedFunction
from pylegend.core.tds.pandas_api.frames.pandas_api_base_tds_frame import PandasApiBaseTdsFrame
from pylegend.core.tds.tds_column import TdsColumn
from pylegend.core.tds.tds_frame import FrameToSqlConfig, FrameToPureConfig
from pylegend.core.tds.cast_helpers import CastTarget, validate_and_build_cast_columns, pure_type_spec, _normalize_target
from pylegend.core.language.shared.helpers import escape_column_name


__all__: PyLegendSequence[str] = [
    "PandasApiCastFunction"
]


class PandasApiCastFunction(PandasApiAppliedFunction):
    __base_frame: PandasApiBaseTdsFrame
    __column_type_map: PyLegendDict[str, CastTarget]

    @classmethod
    def name(cls) -> str:
        return "cast"  # pragma: no cover

    def __init__(
            self,
            base_frame: PandasApiBaseTdsFrame,
            column_type_map: PyLegendDict[str, CastTarget]
    ) -> None:
        self.__base_frame = base_frame
        self.__column_type_map = column_type_map

    def to_sql(self, config: FrameToSqlConfig) -> QuerySpecification:
        return self.__base_frame.to_sql_query_object(config)

    def to_pure(self, config: FrameToPureConfig) -> str:
        base_pure = self.__base_frame.to_pure(config)
        new_columns = validate_and_build_cast_columns(
            self.__base_frame.columns(), self.__column_type_map
        )
        col_specs_parts: PyLegendList[str] = []
        for c in new_columns:
            name = escape_column_name(c.get_name())
            if c.get_name() in self.__column_type_map:
                ptype, params = _normalize_target(self.__column_type_map[c.get_name()])
                col_specs_parts.append(f"{name}:{pure_type_spec(ptype, params)}")
            else:
                col_specs_parts.append(f"{name}:{c.get_type()}")
        col_specs = ", ".join(col_specs_parts)
        return (
            f"{base_pure}{config.separator(1)}"
            f"->cast(@meta::pure::metamodel::relation::Relation<({col_specs})>)"
        )

    def base_frame(self) -> PandasApiBaseTdsFrame:
        return self.__base_frame

    def tds_frame_parameters(self) -> PyLegendList["PandasApiBaseTdsFrame"]:
        return []

    def calculate_columns(self) -> PyLegendSequence["TdsColumn"]:
        return validate_and_build_cast_columns(
            self.__base_frame.columns(), self.__column_type_map
        )

    def validate(self) -> bool:
        validate_and_build_cast_columns(
            self.__base_frame.columns(), self.__column_type_map
        )
        return True



# Copyright 2026 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABCMeta
from pylegend._typing import (
    PyLegendSequence,
    PyLegendDict,
)
from pylegend.core.language.shared.expression import (
    PyLegendExpression,
    PyLegendExpressionBooleanReturn,
    PyLegendExpressionStringReturn,
    PyLegendExpressionNumberReturn,
    PyLegendExpressionIntegerReturn,
    PyLegendExpressionFloatReturn,
    PyLegendExpressionDateReturn,
    PyLegendExpressionDateTimeReturn,
    PyLegendExpressionStrictDateReturn,
)
from pylegend.core.sql.metamodel import (
    Expression,
    QuerySpecification,
)
from pylegend.core.tds.tds_frame import FrameToSqlConfig
from pylegend.core.tds.tds_frame import FrameToPureConfig


__all__: PyLegendSequence[str] = [
    "PyLegendVariableExpression",
    "PyLegendBooleanVariableExpression",
    "PyLegendStringVariableExpression",
    "PyLegendNumberVariableExpression",
    "PyLegendIntegerVariableExpression",
    "PyLegendFloatVariableExpression",
    "PyLegendDateVariableExpression",
    "PyLegendDateTimeVariableExpression",
    "PyLegendStrictDateVariableExpression"
]


class PyLegendVariableExpression(PyLegendExpression, metaclass=ABCMeta):
    __name: str

    def __init__(self, name: str) -> None:
        if not name.isidentifier():
            raise ValueError(f"Invalid variable name: '{name}'. Should be a valid identifier")
        self.__name = name

    def to_sql_expression(
            self,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        raise RuntimeError("SQL translation for variable expression not supported!")

    def to_pure_expression(self, config: FrameToPureConfig) -> str:
        return f"${self.__name}"


class PyLegendBooleanVariableExpression(PyLegendVariableExpression, PyLegendExpressionBooleanReturn):
    def __init__(self, name: str) -> None:
        super().__init__(name=name)


class PyLegendStringVariableExpression(PyLegendVariableExpression, PyLegendExpressionStringReturn):
    def __init__(self, name: str) -> None:
        super().__init__(name=name)


class PyLegendNumberVariableExpression(PyLegendVariableExpression, PyLegendExpressionNumberReturn):
    def __init__(self, name: str) -> None:
        super().__init__(name=name)


class PyLegendIntegerVariableExpression(PyLegendNumberVariableExpression, PyLegendExpressionIntegerReturn):
    def __init__(self, name: str) -> None:
        super().__init__(name=name)


class PyLegendFloatVariableExpression(PyLegendNumberVariableExpression, PyLegendExpressionFloatReturn):
    def __init__(self, name: str) -> None:
        super().__init__(name=name)


class PyLegendDateVariableExpression(PyLegendVariableExpression, PyLegendExpressionDateReturn):
    def __init__(self, name: str) -> None:
        super().__init__(name=name)


class PyLegendDateTimeVariableExpression(PyLegendDateVariableExpression, PyLegendExpressionDateTimeReturn):
    def __init__(self, name: str) -> None:
        super().__init__(name=name)


class PyLegendStrictDateVariableExpression(PyLegendDateVariableExpression, PyLegendExpressionStrictDateReturn):
    def __init__(self, name: str) -> None:
        super().__init__(name=name)



# Copyright 2025 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABCMeta, abstractmethod

from pylegend.core.language import PyLegendBoolean, PyLegendPrimitiveOrPythonPrimitive, PyLegendPrimitiveCollection, \
    PyLegendPrimitive
from pylegend.core.language.legendql_api.legendql_api_custom_expressions import (
    LegendQLApiPrimitive,
    LegendQLApiSortInfo,
    LegendQLApiWindow,
    LegendQLApiPartialFrame,
    LegendQLApiWindowReference,
    LegendQLApiWindowFrame,
)
from pylegend.core.language.legendql_api.legendql_api_tds_row import LegendQLApiTdsRow
from pylegend.core.tds.tds_frame import (
    PyLegendTdsFrame
)
from pylegend._typing import (
    PyLegendSequence,
    PyLegendCallable,
    PyLegendUnion,
    PyLegendList,
    PyLegendOptional,
    PyLegendTuple,
    PyLegendDict,
)
from pylegend.core.tds.cast_helpers import CastTarget

__all__: PyLegendSequence[str] = [
    "LegendQLApiTdsFrame"
]


class LegendQLApiTdsFrame(PyLegendTdsFrame, metaclass=ABCMeta):

    @abstractmethod
    def head(self, count: int = 5) -> "LegendQLApiTdsFrame":
        """
        Return the first n rows of the TDS frame.

        This function returns the first `count` rows from the frame.
        It is useful for quickly inspecting the data without loading the entire dataset.

        Parameters
        ----------
        count : int, default 5
            Number of rows to return.

        Returns
        -------
        LegendQLApiTdsFrame
            A new TDS frame containing only the first n rows.

        See Also
        --------
        limit : Alias for head.
        drop : Skip the first n rows.
        slice : Return a subset of rows by index range.

        Examples
        --------
        .. ipython:: python

            import pylegend
            frame = pylegend.samples.legendql_api.northwind_orders_frame()

            # Get first 5 rows (default)
            frame.head().to_pandas()

            # Get first 3 rows
            frame.head(3).to_pandas()

        """
        pass  # pragma: no cover

    @abstractmethod
    def limit(self, count: int = 5) -> "LegendQLApiTdsFrame":
        """
        Return the first n rows of the TDS frame.

        Alias for :meth:`head`. This function returns the first `count` rows
        from the frame.

        Parameters
        ----------
        count : int, default 5
            Number of rows to return.

        Returns
        -------
        LegendQLApiTdsFrame
            A new TDS frame containing only the first n rows.

        See Also
        --------
        head : Equivalent method.
        drop : Skip the first n rows.

        Examples
        --------
        .. ipython:: python

            import pylegend
            frame = pylegend.samples.legendql_api.northwind_orders_frame()

            # Limit to 2 rows
            frame.limit(2).to_pandas()

        """
        pass  # pragma: no cover

    @abstractmethod
    def drop(self, count: int = 5) -> "LegendQLApiTdsFrame":
        """
        Skip the first n rows of the TDS frame.

        Return a new TDS frame with the first `count` rows removed.
        This is useful for pagination or skipping header-like rows.

        Parameters
        ----------
        count : int, default 5
            Number of rows to skip from the beginning.

        Returns
        -------
        LegendQLApiTdsFrame
            A new TDS frame with the first n rows removed.

        See Also
        --------
        head : Return the first n rows.
        slice : Return a subset of rows by index range.

        Examples
        --------
        .. ipython:: python

            import pylegend
            frame = pylegend.samples.legendql_api.northwind_orders_frame()

            # Skip first 5 rows (default), then show next 5
            frame.drop().head(5).to_pandas()

            # Skip first 10 rows
            frame.drop(10).head(5).to_pandas()

        """
        pass  # pragma: no cover

    @abstractmethod
    def slice(self, start_row: int, end_row_exclusive: int) -> "LegendQLApiTdsFrame":
        """
        Return a subset of rows by index range.

        Select rows from `start_row` (inclusive) to `end_row_exclusive` (exclusive).
        This is similar to Python's list slicing behavior.

        Parameters
        ----------
        start_row : int
            Starting row index (0-based, inclusive).
        end_row_exclusive : int
            Ending row index (0-based, exclusive).

        Returns
        -------
        LegendQLApiTdsFrame
            A new TDS frame containing only the rows in the specified range.

        See Also
        --------
        head : Return the first n rows.
        drop : Skip the first n rows.

        Examples
        --------
        .. ipython:: python

            import pylegend
            frame = pylegend.samples.legendql_api.northwind_orders_frame()

            # Get rows 0-4 (first 5 rows)
            frame.slice(0, 5).to_pandas()

            # Get rows 5-9
            frame.slice(5, 10).to_pandas()

            # Get rows 2-6
            frame.slice(2, 7).to_pandas()

        """
        pass  # pragma: no cover

    @abstractmethod
    def distinct(
            self,
            columns: PyLegendOptional[PyLegendUnion[
                str,
                PyLegendList[str],
                PyLegendCallable[
                    [LegendQLApiTdsRow],
                    PyLegendUnion[LegendQLApiPrimitive, PyLegendList[LegendQLApiPrimitive]]
                ]
            ]] = None
    ) -> "LegendQLApiTdsFrame":
        """
        Return unique rows based on specified columns.

        Remove duplicate rows from the TDS frame. If columns are specified,
        uniqueness is determined based on those columns only. If no columns
        are specified, all columns are used to determine uniqueness.

        Parameters
        ----------
        columns : str, list of str, callable, or None, default None
            Column(s) to consider for identifying duplicates:

            - ``None`` : Use all columns to determine uniqueness.
            - ``str`` : Single column name.
            - ``list of str`` : List of column names.
            - ``callable`` : A function that takes a TDS row and returns
              column(s) to use for uniqueness.

        Returns
        -------
        LegendQLApiTdsFrame
            A new TDS frame with duplicate rows removed.

        See Also
        --------
        select : Select specific columns.
        filter : Filter rows based on a condition.

        Examples
        --------
        .. ipython:: python

            import pylegend
            frame = pylegend.samples.legendql_api.northwind_orders_frame()

            # Get distinct values for a single column
            frame.distinct("Ship Name").head(10).to_pandas()

            # Get distinct combinations of multiple columns
            frame.distinct(["Ship Name", "Order Date"]).head(10).to_pandas()

            # Using a callable to specify columns
            frame.distinct(lambda r: [r["Ship Name"], r["Order Date"]]).head(10).to_pandas()

        """
        pass  # pragma: no cover

    @abstractmethod
    def select(
            self,
            columns: PyLegendUnion[
                str,
                PyLegendList[str],
                PyLegendCallable[
                    [LegendQLApiTdsRow],
                    PyLegendUnion[LegendQLApiPrimitive, PyLegendList[LegendQLApiPrimitive]]
                ]
            ]
    ) -> "LegendQLApiTdsFrame":
        """
        Select specific columns from the TDS frame.

        Return a new TDS frame containing only the specified columns.
        This is useful for reducing the data to only the columns of interest.

        Parameters
        ----------
        columns : str, list of str, or callable
            Column(s) to select:

            - ``str`` : Single column name.
            - ``list of str`` : List of column names.
            - ``callable`` : A function that takes a TDS row and returns
              the column(s) to select.

        Returns
        -------
        LegendQLApiTdsFrame
            A new TDS frame containing only the selected columns.

        See Also
        --------
        distinct : Get unique rows.
        project : Create new computed columns.
        extend : Add new columns while keeping existing ones.

        Examples
        --------
        .. ipython:: python

            import pylegend
            frame = pylegend.samples.legendql_api.northwind_orders_frame()

            # Select a single column
            frame.select("Ship Name").head(5).to_pandas()

            # Select multiple columns
            frame.select(["Order Id", "Ship Name", "Order Date"]).head(5).to_pandas()

            # Using a callable to select columns
            frame.select(lambda r: [r["Order Id"], r["Ship Name"]]).head(5).to_pandas()

        """
        pass  # pragma: no cover

    @abstractmethod
    def sort(
            self,
            sort_infos: PyLegendUnion[
                str,
                PyLegendList[str],
                PyLegendCallable[
                    [LegendQLApiTdsRow],
                    PyLegendUnion[
                        LegendQLApiPrimitive,
                        LegendQLApiSortInfo,
                        PyLegendList[PyLegendUnion[LegendQLApiPrimitive, LegendQLApiSortInfo]],
                    ]
                ]
            ]
    ) -> "LegendQLApiTdsFrame":
        """
        Sort the TDS frame by specified columns.

        Return a new TDS frame sorted by the given column(s). Supports
        ascending and descending sort orders.

        Parameters
        ----------
        sort_infos : str, list of str, or callable
            Column(s) to sort by:

            - ``str`` : Single column name (ascending order).
            - ``list of str`` : List of column names (all ascending order).
            - ``callable`` : A function that takes a TDS row and returns:
                - A column (ascending order by default).
                - A ``LegendQLApiSortInfo`` with direction (use ``.ascending()``
                  or ``.descending()`` on columns).
                - A list of columns and/or ``LegendQLApiSortInfo`` objects.

        Returns
        -------
        LegendQLApiTdsFrame
            A new TDS frame with rows sorted according to the specified criteria.

        See Also
        --------
        head : Return the first n rows.
        filter : Filter rows based on a condition.

        Examples
        --------
        .. ipython:: python

            import pylegend
            frame = pylegend.samples.legendql_api.northwind_orders_frame()

            # Sort by single column (ascending by default)
            frame.sort("Ship Name").head(5).to_pandas()

            # Sort by multiple columns (all ascending)
            frame.sort(["Ship Name", "Order Date"]).head(5).to_pandas()

            # Sort with explicit direction using callable
            frame.sort(lambda r: r["Ship Name"].descending()).head(5).to_pandas()

            # Sort by multiple columns with different directions
            frame.sort(
                lambda r: [r["Ship Name"].ascending(), r["Order Id"].descending()]
            ).head(5).to_pandas()

        """
        pass  # pragma: no cover

    @abstractmethod
    def concatenate(self, other: "LegendQLApiTdsFrame") -> "LegendQLApiTdsFrame":
        """
        Concatenate two TDS frames vertically.

        Append the rows of another TDS frame to the current frame.
        Both frames must have compatible schemas (same columns).

        Parameters
        ----------
        other : LegendQLApiTdsFrame
            The TDS frame to append to the current frame.

        Returns
        -------
        LegendQLApiTdsFrame
            A new TDS frame containing all rows from both frames.

        See Also
        --------
        join : Combine frames horizontally based on a condition.

        Examples
        --------
        .. ipython:: python

            import pylegend
            frame = pylegend.samples.legendql_api.northwind_orders_frame()

            # Get first 3 rows and rows 10-12, then concatenate
            first_rows = frame.head(3)
            other_rows = frame.slice(10, 13)
            first_rows.concatenate(other_rows).to_pandas()

        """
        pass  # pragma: no cover

    @abstractmethod
    def filter(
            self,
            filter_function: PyLegendCallable[[LegendQLApiTdsRow], PyLegendUnion[bool, PyLegendBoolean]]
    ) -> "LegendQLApiTdsFrame":
        """
        Filter rows based on a condition.

        Return a new TDS frame containing only the rows that satisfy
        the given condition.

        Parameters
        ----------
        filter_function : callable
            A function that takes a TDS row and returns a boolean value.
            Rows where the function returns ``True`` are kept.

        Returns
        -------
        LegendQLApiTdsFrame
            A new TDS frame containing only the rows that match the condition.

        See Also
        --------
        distinct : Remove duplicate rows.
        head : Return the first n rows.

        Examples
        --------
        .. ipython:: python

            import pylegend
            frame = pylegend.samples.legendql_api.northwind_orders_frame()

            # Filter by equality
            frame.filter(lambda r: r["Ship Name"] == "Wolski Zajazd").head(5).to_pandas()

            # Filter with comparison operators
            frame.filter(lambda r: r["Order Id"] > 10500).head(5).to_pandas()

            # Filter with multiple conditions (AND)
            frame.filter(
                lambda r: (r["Order Id"] > 10250) & (r["Order Id"] < 10560)
            ).head(5).to_pandas()

        """
        pass  # pragma: no cover

    @abstractmethod
    def rename(
            self,
            column_renames: PyLegendUnion[
                PyLegendTuple[str, str],
                PyLegendList[PyLegendTuple[str, str]],
                PyLegendCallable[
                    [LegendQLApiTdsRow],
                    PyLegendUnion[
                        PyLegendTuple[LegendQLApiPrimitive, str],
                        PyLegendList[PyLegendTuple[LegendQLApiPrimitive, str]]
                    ]
                ]
            ]
    ) -> "LegendQLApiTdsFrame":
        """
        Rename columns in the TDS frame.

        Return a new TDS frame with specified columns renamed.

        Parameters
        ----------
        column_renames : tuple, list of tuples, or callable
            Column rename specification:

            - ``tuple (str, str)`` : Single rename as (old_name, new_name).
            - ``list of tuples`` : Multiple renames as [(old1, new1), (old2, new2), ...].
            - ``callable`` : A function that takes a TDS row and returns
              tuple(s) of (column, new_name).

        Returns
        -------
        LegendQLApiTdsFrame
            A new TDS frame with the specified columns renamed.

        See Also
        --------
        select : Select specific columns.
        extend : Add new columns.
        project : Create new computed columns.

        Examples
        --------
        .. ipython:: python

            import pylegend
            frame = pylegend.samples.legendql_api.northwind_orders_frame()

            # Rename a single column
            frame.rename(("Ship Name", "Vessel Name")).head(3).to_pandas()

            # Rename multiple columns
            frame.rename(
                [("Ship Name", "Vessel Name"), ("Order Date", "Date Ordered")]
            ).head(3).to_pandas()

            # Using a callable to rename columns
            frame.rename(lambda r: (r["Ship Name"], "Vessel Name")).head(3).to_pandas()

        """
        pass  # pragma: no cover

    @abstractmethod
    def extend(
            self,
            extend_columns: PyLegendUnion[
                PyLegendTuple[
                    str,
                    PyLegendCallable[[LegendQLApiTdsRow], PyLegendPrimitiveOrPythonPrimitive]
                ],
                PyLegendTuple[
                    str,
                    PyLegendCallable[[LegendQLApiTdsRow], PyLegendPrimitiveOrPythonPrimitive],
                    PyLegendCallable[[PyLegendPrimitiveCollection], PyLegendPrimitive]
                ],
                PyLegendList[
                    PyLegendUnion[
                        PyLegendTuple[
                            str,
                            PyLegendCallable[[LegendQLApiTdsRow], PyLegendPrimitiveOrPythonPrimitive]
                        ],
                        PyLegendTuple[
                            str,
                            PyLegendCallable[[LegendQLApiTdsRow], PyLegendPrimitiveOrPythonPrimitive],
                            PyLegendCallable[[PyLegendPrimitiveCollection], PyLegendPrimitive]
                        ]
                    ]
                ]
            ]
    ) -> "LegendQLApiTdsFrame":
        """
        Add new computed columns to the TDS frame.

        Return a new TDS frame with additional columns computed from
        existing data. The original columns are preserved.

        Parameters
        ----------
        extend_columns : tuple or list of tuples
            Column extension specification:

            - ``tuple (name, func)`` : Add a column with given name using the
              function to compute values from each row.
            - ``tuple (name, func, agg_func)`` : Add a column with an aggregation
              function applied.
            - ``list of tuples`` : Add multiple columns.

        Returns
        -------
        LegendQLApiTdsFrame
            A new TDS frame with the additional computed columns.

        See Also
        --------
        project : Create new columns (replacing existing).
        rename : Rename existing columns.
        select : Select specific columns.

        Examples
        --------
        .. ipython:: python

            import pylegend
            frame = pylegend.samples.legendql_api.northwind_orders_frame()

            # Add a single computed column
            extended = frame.extend(("Ship Name Upper", lambda r: r["Ship Name"].upper()))
            extended.select(["Ship Name", "Ship Name Upper"]).head(5).to_pandas()

            # Add multiple computed columns
            cols = ["Ship Name", "Ship Name Upper", "Ship Name Lower"]
            frame.extend([
                ("Ship Name Upper", lambda r: r["Ship Name"].upper()),
                ("Ship Name Lower", lambda r: r["Ship Name"].lower())
            ]).select(cols).head(5).to_pandas()

        """
        pass  # pragma: no cover

    @abstractmethod
    def join(
            self,
            other: "LegendQLApiTdsFrame",
            join_condition: PyLegendCallable[
                [LegendQLApiTdsRow, LegendQLApiTdsRow], PyLegendUnion[bool, PyLegendBoolean]
            ],
            join_type: str = 'LEFT_OUTER'
    ) -> "LegendQLApiTdsFrame":
        """
        Join two TDS frames based on a condition.

        Combine columns from two TDS frames based on a join condition.

        Parameters
        ----------
        other : LegendQLApiTdsFrame
            The right TDS frame to join with.
        join_condition : callable
            A function that takes two TDS rows (left, right) and returns
            a boolean indicating whether the rows should be joined.
        join_type : str, default 'LEFT_OUTER'
            Type of join to perform:

            - ``'INNER'`` : Only matching rows from both frames.
            - ``'LEFT_OUTER'`` : All rows from left, matching from right.
            - ``'RIGHT_OUTER'`` : All rows from right, matching from left.

        Returns
        -------
        LegendQLApiTdsFrame
            A new TDS frame with columns from both frames.

        See Also
        --------
        inner_join : Convenience method for inner join.
        left_join : Convenience method for left outer join.
        right_join : Convenience method for right outer join.
        as_of_join : Join based on temporal proximity.
        concatenate : Combine frames vertically.

        Examples
        --------
        .. ipython:: python

            import pylegend
            orders = pylegend.samples.legendql_api.northwind_orders_frame()

            # Self-join example: join orders with a filtered version
            # Rename duplicate columns to avoid conflicts
            cols = ["Order Id", "Ship Name"]
            renames = [("Order Id", "Right Order Id"), ("Ship Name", "Filtered Ship Name")]
            filtered_orders = orders.filter(
                lambda r: r["Order Id"] > 10300
            ).select(cols).rename(renames)
            orders.join(
                filtered_orders,
                lambda l, r: l["Order Id"] == r["Right Order Id"],
                "INNER"
            ).head(5).to_pandas()

        """
        pass  # pragma: no cover

    @abstractmethod
    def inner_join(
            self,
            other: "LegendQLApiTdsFrame",
            join_condition: PyLegendCallable[
                [LegendQLApiTdsRow, LegendQLApiTdsRow], PyLegendUnion[bool, PyLegendBoolean]
            ]
    ) -> "LegendQLApiTdsFrame":
        """
        Perform an inner join with another TDS frame.

        Return only the rows that have matching values in both frames.

        Parameters
        ----------
        other : LegendQLApiTdsFrame
            The right TDS frame to join with.
        join_condition : callable
            A function that takes two TDS rows (left, right) and returns
            a boolean indicating whether the rows should be joined.

        Returns
        -------
        LegendQLApiTdsFrame
            A new TDS frame containing only matching rows from both frames.

        See Also
        --------
        join : General join method with configurable join type.
        left_join : Left outer join.
        right_join : Right outer join.

        Examples
        --------
        .. ipython:: python

            import pylegend
            orders = pylegend.samples.legendql_api.northwind_orders_frame()

            # Inner join with filtered data
            # Rename duplicate columns to avoid conflicts
            cols = ["Order Id", "Ship Name"]
            renames = [("Order Id", "Right Order Id"), ("Ship Name", "Filtered Ship Name")]
            filtered = orders.filter(
                lambda r: r["Order Id"] > 10300
            ).select(cols).rename(renames)
            orders.inner_join(
                filtered,
                lambda l, r: l["Order Id"] == r["Right Order Id"]
            ).head(5).to_pandas()

        """
        pass  # pragma: no cover

    @abstractmethod
    def left_join(
            self,
            other: "LegendQLApiTdsFrame",
            join_condition: PyLegendCallable[
                [LegendQLApiTdsRow, LegendQLApiTdsRow], PyLegendUnion[bool, PyLegendBoolean]
            ]
    ) -> "LegendQLApiTdsFrame":
        """
        Perform a left outer join with another TDS frame.

        Return all rows from the left frame, with matching rows from the
        right frame. Non-matching rows will have null values for the
        right frame columns.

        Parameters
        ----------
        other : LegendQLApiTdsFrame
            The right TDS frame to join with.
        join_condition : callable
            A function that takes two TDS rows (left, right) and returns
            a boolean indicating whether the rows should be joined.

        Returns
        -------
        LegendQLApiTdsFrame
            A new TDS frame with all left rows and matching right rows.

        See Also
        --------
        join : General join method with configurable join type.
        inner_join : Inner join.
        right_join : Right outer join.

        Examples
        --------
        .. ipython:: python

            import pylegend
            orders = pylegend.samples.legendql_api.northwind_orders_frame()

            # Left join - all orders with filtered info where available
            # Rename duplicate columns to avoid conflicts
            cols = ["Order Id", "Ship Name"]
            renames = [("Order Id", "Right Order Id"), ("Ship Name", "Filtered Ship Name")]
            filtered_info = orders.filter(
                lambda r: r["Order Id"] > 10300
            ).select(cols).rename(renames).head(10)
            orders.head(15).left_join(
                filtered_info,
                lambda l, r: l["Order Id"] == r["Right Order Id"]
            ).to_pandas()

        """
        pass  # pragma: no cover

    @abstractmethod
    def right_join(
            self,
            other: "LegendQLApiTdsFrame",
            join_condition: PyLegendCallable[
                [LegendQLApiTdsRow, LegendQLApiTdsRow], PyLegendUnion[bool, PyLegendBoolean]
            ]
    ) -> "LegendQLApiTdsFrame":
        """
        Perform a right outer join with another TDS frame.

        Return all rows from the right frame, with matching rows from the
        left frame. Non-matching rows will have null values for the
        left frame columns.

        Parameters
        ----------
        other : LegendQLApiTdsFrame
            The right TDS frame to join with.
        join_condition : callable
            A function that takes two TDS rows (left, right) and returns
            a boolean indicating whether the rows should be joined.

        Returns
        -------
        LegendQLApiTdsFrame
            A new TDS frame with all right rows and matching left rows.

        See Also
        --------
        join : General join method with configurable join type.
        inner_join : Inner join.
        left_join : Left outer join.

        Examples
        --------
        .. ipython:: python

            import pylegend
            orders = pylegend.samples.legendql_api.northwind_orders_frame()

            # Right join example
            # Rename duplicate columns to avoid conflicts
            cols = ["Order Id", "Ship Name"]
            renames = [("Order Id", "Right Order Id"), ("Ship Name", "Filtered Ship Name")]
            filtered_info = orders.filter(
                lambda r: r["Order Id"] > 10300
            ).select(cols).rename(renames).head(10)
            orders.head(5).right_join(
                filtered_info,
                lambda l, r: l["Order Id"] == r["Right Order Id"]
            ).to_pandas()

        """
        pass  # pragma: no cover

    @abstractmethod
    def as_of_join(
            self,
            other: "LegendQLApiTdsFrame",
            match_function: PyLegendCallable[
                [LegendQLApiTdsRow, LegendQLApiTdsRow], PyLegendUnion[bool, PyLegendBoolean]
            ],
            join_condition: PyLegendOptional[
                PyLegendCallable[[LegendQLApiTdsRow, LegendQLApiTdsRow], PyLegendUnion[bool, PyLegendBoolean]]
            ] = None
    ) -> "LegendQLApiTdsFrame":
        pass  # pragma: no cover

    @abstractmethod
    def group_by(
            self,
            grouping_columns: PyLegendUnion[
                str,
                PyLegendList[str],
                PyLegendCallable[
                    [LegendQLApiTdsRow],
                    PyLegendUnion[LegendQLApiPrimitive, PyLegendList[LegendQLApiPrimitive]]
                ]
            ],
            aggregate_specifications: PyLegendUnion[
                PyLegendTuple[
                    str,
                    PyLegendCallable[[LegendQLApiTdsRow], PyLegendPrimitiveOrPythonPrimitive],
                    PyLegendCallable[[PyLegendPrimitiveCollection], PyLegendPrimitive]
                ],
                PyLegendList[
                    PyLegendTuple[
                        str,
                        PyLegendCallable[[LegendQLApiTdsRow], PyLegendPrimitiveOrPythonPrimitive],
                        PyLegendCallable[[PyLegendPrimitiveCollection], PyLegendPrimitive]
                    ]
                ]
            ]
    ) -> "LegendQLApiTdsFrame":
        """
        Group rows and apply aggregate functions.

        Group the TDS frame by the specified columns and compute
        aggregate values for each group.

        Parameters
        ----------
        grouping_columns : str, list of str, or callable
            Column(s) to group by:

            - ``str`` : Single column name.
            - ``list of str`` : Multiple column names.
            - ``callable`` : A function returning column(s) to group by.

        aggregate_specifications : tuple or list of tuples
            Aggregation specification as tuple(s) of:
            ``(result_column_name, column_selector, aggregate_function)``

            Common aggregate functions from ``pylegend.agg``:

            - ``count()`` : Count of values.
            - ``sum()`` : Sum of values.
            - ``avg()`` : Average of values.
            - ``min()`` : Minimum value.
            - ``max()`` : Maximum value.

        Returns
        -------
        LegendQLApiTdsFrame
            A new TDS frame with grouped and aggregated data.

        See Also
        --------
        distinct : Get unique rows.
        filter : Filter rows before grouping.

        Examples
        --------
        .. ipython:: python

            import pylegend
            frame = pylegend.samples.legendql_api.northwind_orders_frame()

            # Group by single column with count using lambda syntax
            frame.group_by(
                lambda r: r["Ship Name"],
                ("Order Count", lambda r: r["Order Id"], lambda col: col.count())
            ).head(10).to_pandas()

            # Group by single column with multiple aggregations using lambda syntax
            frame.group_by(
                lambda r: r["Ship Name"],
                [
                    ("Order Count", lambda r: r["Order Id"], lambda col: col.count()),
                    ("Min Order Id", lambda r: r["Order Id"], lambda col: col.min()),
                    ("Max Order Id", lambda r: r["Order Id"], lambda col: col.max())
                ]
            ).head(10).to_pandas()

            # Chained group_by operations - grouping on result of previous aggregation
            frame2 = frame.group_by(
                lambda r: r["Ship Name"],
                ("Order Count", lambda r: r["Order Id"], lambda col: col.count())
            )
            frame2.group_by(
                lambda r: r["Ship Name"],
                ("Count of Counts", lambda r: r["Order Count"], lambda col: col.count())
            ).head(10).to_pandas()

        """
        pass  # pragma: no cover

    @abstractmethod
    def window(
            self,
            partition_by: PyLegendOptional[
                PyLegendUnion[
                    str,
                    PyLegendList[str],
                    PyLegendCallable[
                        [LegendQLApiTdsRow],
                        PyLegendUnion[LegendQLApiPrimitive, PyLegendList[LegendQLApiPrimitive]]
                    ]
                ]
            ] = None,
            order_by: PyLegendOptional[
                PyLegendUnion[
                    str,
                    PyLegendList[str],
                    PyLegendCallable[
                        [LegendQLApiTdsRow],
                        PyLegendUnion[
                            LegendQLApiPrimitive,
                            LegendQLApiSortInfo,
                            PyLegendList[PyLegendUnion[LegendQLApiPrimitive, LegendQLApiSortInfo]],
                        ]
                    ]
                ]
            ] = None,
            frame: PyLegendOptional[LegendQLApiWindowFrame] = None
    ) -> LegendQLApiWindow:
        """
        Create a window specification for window functions.

        Define how rows are partitioned, ordered, and framed for
        window function calculations.

        Parameters
        ----------
        partition_by : str, list of str, callable, or None, optional
            Column(s) to partition the data by. Each partition is processed
            independently for window calculations.

            - ``None`` : Entire dataset is one partition.
            - ``str`` : Single column name.
            - ``list of str`` : Multiple column names.
            - ``callable`` : Function returning column(s).

        order_by : str, list of str, callable, or None, optional
            Column(s) to order rows within each partition.

            - ``None`` : No specific order.
            - ``str`` : Single column name (ascending).
            - ``list of str`` : Multiple columns (ascending).
            - ``callable`` : Function returning column(s) with optional
              ``.ascending()`` or ``.descending()`` direction.

        frame : LegendQLApiWindowFrame, optional
            Window frame specification created by :meth:`rows` or :meth:`range`.
            Defines which rows relative to the current row are included
            in the calculation.

        Returns
        -------
        LegendQLApiWindow
            A window specification for use with :meth:`window_extend`.

        See Also
        --------
        rows : Create a row-based window frame.
        range : Create a range-based window frame.
        window_extend : Apply window functions using the window specification.

        """
        pass

    @abstractmethod
    def window_extend(
            self,
            window: LegendQLApiWindow,
            extend_columns: PyLegendUnion[
                PyLegendTuple[
                    str,
                    PyLegendCallable[
                        [LegendQLApiPartialFrame, LegendQLApiWindowReference, LegendQLApiTdsRow],
                        PyLegendPrimitiveOrPythonPrimitive
                    ]
                ],
                PyLegendTuple[
                    str,
                    PyLegendCallable[
                        [LegendQLApiPartialFrame, LegendQLApiWindowReference, LegendQLApiTdsRow],
                        PyLegendPrimitiveOrPythonPrimitive
                    ],
                    PyLegendCallable[[PyLegendPrimitiveCollection], PyLegendPrimitive]
                ],
                PyLegendList[
                    PyLegendUnion[
                        PyLegendTuple[
                            str,
                            PyLegendCallable[
                                [LegendQLApiPartialFrame, LegendQLApiWindowReference, LegendQLApiTdsRow],
                                PyLegendPrimitiveOrPythonPrimitive
                            ]
                        ],
                        PyLegendTuple[
                            str,
                            PyLegendCallable[
                                [LegendQLApiPartialFrame, LegendQLApiWindowReference, LegendQLApiTdsRow],
                                PyLegendPrimitiveOrPythonPrimitive
                            ],
                            PyLegendCallable[[PyLegendPrimitiveCollection], PyLegendPrimitive]
                        ]
                    ]
                ]
            ]
    ) -> "LegendQLApiTdsFrame":
        """
        Add columns computed using window functions.

        Extend the TDS frame with new columns calculated using window
        functions over the specified window.

        Parameters
        ----------
        window : LegendQLApiWindow
            Window specification created by :meth:`window`.
        extend_columns : tuple or list of tuples
            Column specification(s) as:

            - ``tuple (name, func)`` : Column name and function that receives
              (partial_frame, window_ref, row) and returns a value.
            - ``tuple (name, func, agg_func)`` : With aggregation function.
            - ``list of tuples`` : Multiple window columns.

            The function parameters:

            - ``partial_frame`` : Access to partitioned frame data.
            - ``window_ref`` : Window reference for functions like ``row_number()``,
              ``rank()``, ``lead()``, ``lag()``.
            - ``row`` : Current row data.

        Returns
        -------
        LegendQLApiTdsFrame
            A new TDS frame with additional window-computed columns.

        See Also
        --------
        window : Create a window specification.
        rows : Create a row-based window frame.
        range : Create a range-based window frame.
        extend : Add computed columns without window functions.

        Examples
        --------
        .. ipython:: python

            import pylegend
            frame = pylegend.samples.legendql_api.northwind_orders_frame()

            # Add a single window function column
            win = frame.window(partition_by="Ship Name", order_by="Order Id")
            frame.window_extend(
                win,
                ("RowNumber", lambda p, w, r: p.row_number(r))
            ).select(
                ["Ship Name", "Order Id", "RowNumber"]
            ).head(5).to_pandas()

            # Add basic window function columns
            frame.window_extend(
                frame.window(partition_by="Ship Name", order_by="Order Id"),
                [
                    ("RowNumber", lambda p, w, r: p.row_number(r)),
                    ("Rank", lambda p, w, r: p.rank(w, r))
                ]
            ).select(
               ["Ship Name", "Order Id", "RowNumber", "Rank"]
            ).head(5).to_pandas()

        """
        pass

    @abstractmethod
    def project(
            self,
            project_columns: PyLegendUnion[
                PyLegendTuple[
                    str,
                    PyLegendCallable[[LegendQLApiTdsRow], PyLegendPrimitiveOrPythonPrimitive]
                ],
                PyLegendList[
                    PyLegendTuple[
                        str,
                        PyLegendCallable[[LegendQLApiTdsRow], PyLegendPrimitiveOrPythonPrimitive]
                    ]
                ]
            ]
    ) -> "LegendQLApiTdsFrame":
        """
        Create a new TDS frame with specified computed columns.

        Unlike :meth:`extend`, which adds columns to existing ones,
        ``project`` creates a new frame with only the specified columns.

        Parameters
        ----------
        project_columns : tuple or list of tuples
            Column specification(s) as ``(column_name, function)``:

            - ``tuple (str, callable)`` : Single column with name and
              computation function.
            - ``list of tuples`` : Multiple columns.

            The function takes a TDS row and returns the column value.

        Returns
        -------
        LegendQLApiTdsFrame
            A new TDS frame containing only the projected columns.

        See Also
        --------
        extend : Add columns while keeping existing ones.
        select : Select existing columns.
        rename : Rename existing columns.

        Examples
        --------
        .. ipython:: python

            import pylegend
            frame = pylegend.samples.legendql_api.northwind_orders_frame()

            # Project single column (copy with new name)
            frame.project(("Order Number", lambda r: r["Order Id"])).head(5).to_pandas()

            # Project multiple columns with transformations
            frame.project([
                ("Order Number", lambda r: r["Order Id"]),
                ("Vessel", lambda r: r["Ship Name"].upper()),
                ("Vessel Lower", lambda r: r["Ship Name"].lower())
            ]).head(5).to_pandas()

            # Project with existing column values
            frame.project([
                ("ID", lambda r: r["Order Id"]),
                ("Name", lambda r: r["Ship Name"]),
                ("Date", lambda r: r["Order Date"])
            ]).head(5).to_pandas()

        """
        pass  # pragma: no cover

    @abstractmethod
    def rows(
            self,
            start: PyLegendUnion[str, int],
            end: PyLegendUnion[str, int]) -> LegendQLApiWindowFrame:
        """
        Create a row-based window frame specification.

        Define a window frame based on row offsets relative to the current row.
        Used with :meth:`window` and :meth:`window_extend` for window functions.

        Parameters
        ----------
        start : str or int
            Start boundary of the window:

            - ``"unbounded"`` : From the first row of the partition.
            - ``0`` : Current row.
            - Negative int : That many rows before current row.
            - Positive int : That many rows after current row.

        end : str or int
            End boundary of the window:

            - ``"unbounded"`` : To the last row of the partition.
            - ``0`` : Current row.
            - Negative int : That many rows before current row.
            - Positive int : That many rows after current row.

        Returns
        -------
        LegendQLApiWindowFrame
            A window frame specification for use with window functions.

        See Also
        --------
        range : Create a range-based window frame.
        window : Create a window specification.
        window_extend : Apply window functions.

        """
        pass  # pragma: no cover

    @abstractmethod
    def range(
            self,
            *,
            number_start: PyLegendOptional[PyLegendUnion[str, int, float]] = None,
            number_end: PyLegendOptional[PyLegendUnion[str, int, float]] = None,
            duration_start: PyLegendOptional[PyLegendUnion[str, int, float]] = None,
            duration_start_unit: PyLegendOptional[str] = None,
            duration_end: PyLegendOptional[PyLegendUnion[str, int, float]] = None,
            duration_end_unit: PyLegendOptional[str] = None) -> LegendQLApiWindowFrame:
        """
        Create a range-based window frame specification.

        Define a window frame based on value ranges (numeric or duration)
        relative to the current row's order-by column value.
        Used with :meth:`window` and :meth:`window_extend` for window functions.

        Parameters
        ----------
        number_start : str, int, or float, optional
            Numeric start boundary. Use ``"unbounded"`` for no lower limit.
        number_end : str, int, or float, optional
            Numeric end boundary. Use ``"unbounded"`` for no upper limit.
        duration_start : str, int, or float, optional
            Duration start boundary for time-based ranges.
        duration_start_unit : str, optional
            Time unit for duration_start (e.g., ``"DAYS"``, ``"HOURS"``).
        duration_end : str, int, or float, optional
            Duration end boundary for time-based ranges.
        duration_end_unit : str, optional
            Time unit for duration_end.

        Returns
        -------
        LegendQLApiWindowFrame
            A window frame specification for use with window functions.

        Raises
        ------
        ValueError
            If neither numeric nor duration range is provided, or if both are provided.

        See Also
        --------
        rows : Create a row-based window frame.
        window : Create a window specification.
        window_extend : Apply window functions.

        """
        pass  # pragma: no cover

    @abstractmethod
    def cast(
            self,
            column_type_map: PyLegendDict[str, CastTarget]
    ) -> "LegendQLApiTdsFrame":
        """
        Cast specific columns to new types.

        Return a new TDS frame where the types of the specified columns
        are updated according to the provided type map. All other columns
        retain their original types.

        Parameters
        ----------
        column_type_map : dict
            A mapping from column names to target types. Use functions from
            the `pylegend.type_factory` module to create target types.

        Returns
        -------
        LegendQLApiTdsFrame
            A new TDS frame with the specified columns cast to the new types.

        Examples
        --------
        .. ipython:: python

            from pylegend import type_factory as tf
            from pylegend.core.tds.tds_column import PrimitiveTdsColumn
            from pylegend.extensions.tds.legendql_api.frames.legendql_api_table_spec_input_frame import (
                LegendQLApiTableSpecInputFrame
            )

            frame = LegendQLApiTableSpecInputFrame(
                ['db', 'schema', 'order'],
                [
                    PrimitiveTdsColumn.integer_column("Order Id"),
                    PrimitiveTdsColumn.date_column("Order Date"),
                    PrimitiveTdsColumn.string_column("Ship Name"),
                    PrimitiveTdsColumn.float_column("Amount"),
                    PrimitiveTdsColumn.decimal_column("Discount"),
                ]
            )

            # Cast using simple types
            cast_frame_1 = frame.cast({
                "Order Date": tf.datetime(),
                "Order Id": tf.number(),
                "Amount": tf.number(),
            })
            print(cast_frame_1.to_pure_query())

            # Cast using precise primitive types (including parameterized)
            cast_frame_2 = frame.cast({
                "Order Date": tf.timestamp(),
                "Order Id": tf.bigint(),
                "Ship Name": tf.varchar(200),
                "Amount": tf.float4(),
                "Discount": tf.numeric(10, 2),
            })
            print(cast_frame_2.to_pure_query())

        """
        pass  # pragma: no cover



# Copyright 2023 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import date
from pylegend._typing import (
    PyLegendSequence,
    PyLegendDict,
    PyLegendUnion,
)
from pylegend.core.language.shared.primitives.date import PyLegendDate
from pylegend.core.language.shared.expression import (
    PyLegendExpressionStrictDateReturn,
)
from pylegend.core.language.shared.literal_expressions import (
    PyLegendStrictDateLiteralExpression,
    PyLegendIntegerLiteralExpression,
    PyLegendStringLiteralExpression
)
from pylegend.core.sql.metamodel import (
    Expression,
    QuerySpecification
)
from pylegend.core.tds.pandas_api.frames.helpers.series_helper import grammar_method
from pylegend.core.tds.tds_frame import FrameToSqlConfig
from pylegend.core.language.shared.operations.date_operation_expressions import PyLegendDateTimeBucketExpression
from pylegend.core.language.shared.primitives.integer import PyLegendInteger

__all__: PyLegendSequence[str] = [
    "PyLegendStrictDate"
]


class PyLegendStrictDate(PyLegendDate):
    __value: PyLegendExpressionStrictDateReturn

    def __init__(
            self,
            value: PyLegendExpressionStrictDateReturn
    ) -> None:
        super().__init__(value)
        self.__value = value

    def to_sql_expression(
            self,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return self.__value.to_sql_expression(frame_name_to_base_query_map, config)

    def value(self) -> PyLegendExpressionStrictDateReturn:
        return self.__value

    @grammar_method
    def time_bucket(
            self,
            quantity: PyLegendUnion[int, "PyLegendInteger"],
            duration_unit: str) -> "PyLegendDate":
        self.validate_param_to_be_int_or_int_expr(quantity, "time bucket quantity parameter")
        quantity_op = PyLegendIntegerLiteralExpression(quantity) if isinstance(quantity, int) else quantity.value()
        self.validate_duration_unit_param(duration_unit)
        duration_unit_op = PyLegendStringLiteralExpression(duration_unit.upper())
        return PyLegendDate(PyLegendDateTimeBucketExpression([
            self.__value,
            quantity_op,
            duration_unit_op,
            PyLegendStringLiteralExpression("STRICTDATE")]))

    @staticmethod
    def __convert_to_strictdate_expr(
            val: PyLegendUnion[date, "PyLegendStrictDate"]
    ) -> PyLegendExpressionStrictDateReturn:
        if isinstance(val, date):
            return PyLegendStrictDateLiteralExpression(val)
        return val.__value

    @staticmethod
    def validate_param_to_be_strictdate(
            param: PyLegendUnion[date, "PyLegendStrictDate"],
            desc: str
    ) -> None:
        if not isinstance(param, (date, PyLegendStrictDate)):
            raise TypeError(desc + " should be a datetime.date or a StrictDate expression (PyLegendStrictDate)."
                                   " Got value " + str(param) + " of type: " + str(type(param)))

    @staticmethod
    def validate_duration_unit_param(duration_unit: str) -> None:
        if duration_unit.lower() not in ('years', 'months', 'weeks', 'days'):
            raise ValueError(
                f"Unknown duration unit - {duration_unit}. Supported values are - YEARS, MONTHS, WEEKS, DAYS"
            )
